/**
 * AbstractPlotWindow -- Partial implementation of a plot window.
 *
 * Copyright (C) 2011-2024, by Joseph A. Huwaldt. All rights reserved.
 *
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation;
 * either version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
 * PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple Place -
 * Suite 330, Boston, MA 02111-1307, USA. Or visit: http://www.gnu.org/licenses/lgpl.html
 */

import jahuwaldt.io.ExtFilenameFilter;
import jahuwaldt.swing.*;
import java.awt.Dimension;
import java.awt.Graphics2D;
import java.awt.Image;
import java.awt.Toolkit;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.WindowEvent;
import java.awt.geom.Rectangle2D;
import java.awt.print.PageFormat;
import java.awt.print.PrinterException;
import java.awt.print.PrinterJob;
import java.io.*;
import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.text.MessageFormat;
import java.util.ResourceBundle;
import java.util.logging.Level;
import java.util.logging.Logger;
import javax.swing.*;
import org.jfree.chart.ChartPanel;
import org.jfree.chart.ChartUtilities;
import org.jfree.chart.JFreeChart;

/**
 * A partial implementation of a window that displays a plot using JFreeChart.
 *
 * <p> Modified by: Joseph A. Huwaldt </p>
 *
 * @author Joseph A. Huwaldt Date: September 3, 2011
 * @version January 15, 2024
 */
public abstract class AbstractPlotWindow extends JFrame {

    //  The logger for this class.
    private static final Logger LOGGER = Logger.getLogger(AbstractPlotWindow.class.getName());
    
    private static final long serialVersionUID = 1L;

    private static final short FRAME_WIDTH = 1020;   // Application frame width & height.
    private static final short FRAME_HEIGHT = 640;

    //  The name of the plot.
    private final String _dataName;

    //  The page format used for printing.
    private PageFormat _pf = null;

    //  The panel used to display the chart.
    private ChartPanel _chartPanel;

    /**
     * Construct a new AbstractPlotWindow with the given parameters.
     *
     * @param title    The title for the window.
     * @param dataName Name of the data being plotted (possibly the name of the file it
     *                 came from).
     * @throws java.lang.NoSuchMethodException if there is any problem creating the
     * window.
     */
    @SuppressWarnings({"LeakingThisInConstructor", "OverridableMethodCallInConstructor"})
    public AbstractPlotWindow(String title, String dataName) throws NoSuchMethodException {
        super(title);

        MDIApplication.addWindow(this);
        _dataName = dataName;

        //  Define the window icon in the taskbar.
        ImageIcon icon = AppUtilities.getResourceIcon(MDIApplication.getInstance().getResourceBundle().getString("applicationIconResName"), LOGGER);
        if (icon != null)
            this.setIconImage(icon.getImage());

        setResizable(true);
        setSize(FRAME_WIDTH, FRAME_HEIGHT);

        //  Position the window so that each new window can be seen.
        AppUtilities.positionWindow(this, FRAME_WIDTH, FRAME_HEIGHT);

        //  Have the window dispose of itself when it closes.
        setDefaultCloseOperation(WindowConstants.DISPOSE_ON_CLOSE);

        //  Create a menu bar for this window.
        JMenuBar menuBar = createMenuBar();
        this.setJMenuBar(menuBar);

    }

    /**
     * Sets the title for this window to the specified string. Also notifies the main
     * application class so that the "Windows" menu gets updated too.
     *
     * @param title The title to be displayed in the frame's border. A null value is
     *              treated as an empty string, "".
     */
    @Override
    public void setTitle(String title) {
        super.setTitle(title);
        MDIApplication.windowTitleChanged(this);
    }

    /**
     * Return the name to use for the data in this window.
     *
     * @return The name for the data in this window.
     */
    public String getDataName() {
        return _dataName;
    }

    /**
     * Initializes the menus associated with this window.
     */
    private JMenuBar createMenuBar() throws NoSuchMethodException {
        MDIApplication guiApp = MDIApplication.getInstance();
        ResourceBundle resBundle = guiApp.getResourceBundle();

        JMenuBar menuBar = new JMenuBar();

        // Set up the file menu.
        int row = 0;
        String[][] menuStrings = new String[11][3];
        menuStrings[row][0] = resBundle.getString("newItemText");
        menuStrings[row][1] = resBundle.getString("newItemKey");
        menuStrings[row][2] = "handleNew";
        ++row;
        menuStrings[row][0] = resBundle.getString("openItemText");
        menuStrings[row][1] = resBundle.getString("openItemKey");
        menuStrings[row][2] = null;
        ++row;
        ++row;  //  Blank line
        menuStrings[row][0] = resBundle.getString("closeItemText");
        menuStrings[row][1] = resBundle.getString("closeItemKey");
        menuStrings[row][2] = "handleClose";
        ++row;
        menuStrings[row][0] = resBundle.getString("saveItemText");
        menuStrings[row][1] = resBundle.getString("saveItemKey");
        menuStrings[row][2] = null;
        ++row;
        menuStrings[row][0] = resBundle.getString("saveAsPNGItemText");
        menuStrings[row][1] = null;
        menuStrings[row][2] = "handleSaveAsPNG";
        ++row;
        if (createSVGGraphics2D(10, 10) != null) {
            menuStrings[row][0] = resBundle.getString("saveAsSVGItemText");
            menuStrings[row][1] = null;
            menuStrings[row][2] = "handleSaveAsSVG";
            ++row;
        }
        int exportAsMItem = row;
        menuStrings[row][0] = resBundle.getString("exportAsItemText");
        menuStrings[row][1] = null;
        menuStrings[row][2] = null;
        ++row;
        ++row;  //  Blank line
        menuStrings[row][0] = resBundle.getString("pageSetupItemText");
        menuStrings[row][1] = null;
        menuStrings[row][2] = "handlePageSetup";
        ++row;
        menuStrings[row][0] = resBundle.getString("printItemText");
        menuStrings[row][1] = resBundle.getString("printItemKey");
        menuStrings[row][2] = "handlePrint";
        JMenu menu = AppUtilities.buildMenu(this, resBundle.getString("fileMenuText"), menuStrings);
        menuBar.add(menu);

        //  Add the "Export As" sub-menu.
        JMenu exportMenu = buildExportMenu(menu.getItem(exportAsMItem).getText());
        if (exportMenu != null) {
            menu.remove(exportAsMItem);
            menu.add(exportMenu, exportAsMItem);
        }

        //  Add a Quit menu item (if it isn't automatically present).
        if (!QuitJMenuItem.isAutomaticallyPresent()) {
            menu.addSeparator();
            menu.add(guiApp.getQuitJMenuItem());
        }


        // Set up the edit menu.
        row = 0;
        menuStrings = new String[6][3];
        menuStrings[row][0] = resBundle.getString("undoItemText");
        menuStrings[row][1] = resBundle.getString("undoItemKey");
        menuStrings[row][2] = null;
        ++row;
        menuStrings[row][0] = resBundle.getString("redoItemText");
        menuStrings[row][1] = null;
        menuStrings[row][2] = null;
        ++row;
        ++row;  //  Blank line.
        menuStrings[row][0] = resBundle.getString("cutItemText");
        menuStrings[row][1] = resBundle.getString("cutItemKey");
        menuStrings[row][2] = null;
        ++row;
        menuStrings[row][0] = resBundle.getString("copyItemText");
        menuStrings[row][1] = resBundle.getString("copyItemKey");
        menuStrings[row][2] = "handleCopy";
        ++row;
        menuStrings[row][0] = resBundle.getString("pasteItemText");
        menuStrings[row][1] = resBundle.getString("pasteItemKey");
        menuStrings[row][2] = null;
        menu = AppUtilities.buildMenu(this, resBundle.getString("editMenuText"), menuStrings);
        menuBar.add(menu);

        //  Create a Window's menu.
        menu = MDIApplication.newWindowsMenu(resBundle.getString("windowsMenuText"));
        menuBar.add(menu);

        //  Create an about menu item (if not automatically present).
        if (!AboutJMenuItem.isAutomaticallyPresent()) {
            //  Create Help menu.
            menu = new JMenu(resBundle.getString("helpMenuText"));
            menuBar.add(menu);

            //  Add the "About" item to the Help menu.
            menu.add(guiApp.getAboutJMenuItem());
        }

        return menuBar;
    }

    /**
     * Method that will build an "Export" menu containing a list of all the objects that
     * are capable of writing to files.
     *
     * @param title The title for the menu that is created.
     * @return A JMenu instance containing a list of all the objects that can write to
     * files in various formats. If there are no appropriate readers available, null is
     * returned.
     * @throws java.lang.NoSuchMethodException if there is any problem building the export
     * menu.
     */
    protected abstract JMenu buildExportMenu(String title) throws NoSuchMethodException;

    /**
     * Return a reference to the panel used to display the chart (may be <code>null</code>
     * if the chart has not been created yet).
     *
     * @return a reference to the chart panel.
     */
    public ChartPanel getChartPanel() {
        return _chartPanel;
    }

    /**
     * Creates a panel that contains the chart of the data and modifies the chart's pop-up
     * menu to provide custom behavior.
     *
     * @param chart  The chart to add to the chart panel.
     * @param target The object that receives the action events. Must have the following
     *               methods defined: "handlePrint(ActionEvent e)",
     *               "handlePageSetup(ActionEvent e)", "handleSaveAsPNG(ActionEvent e)",
     *               and "handleSaveAsSVG(ActionEvent e)".
     * @throws java.lang.NoSuchMethodException if there is any problem creating the chart
     * panel.
     * @see getChartPanel
     */
    protected void createChartPanel(JFreeChart chart, Object target) throws NoSuchMethodException {

        _chartPanel = new ChartPanel(chart);
        _chartPanel.setMouseZoomable(true);
        _chartPanel.setFillZoomRectangle(false);    // Use an outline (performance issues on Windows with filled rectangle).
        _chartPanel.setHorizontalAxisTrace(false);
        _chartPanel.setVerticalAxisTrace(false);
        _chartPanel.setPreferredSize(new Dimension(640, 480));

        //  Modify the chart panel's pop-up menu.
        JPopupMenu chartMenu = _chartPanel.getPopupMenu();

        //  Replace the default Print... menu item with our own.
        chartMenu.remove(5);

        //  Add our own Print... item to chart menu.
        ResourceBundle resBundle = MDIApplication.getInstance().getResourceBundle();
        JMenuItem item = new JMenuItem(resBundle.getString("printItemText"));
        ActionListener listener = AppUtilities.getActionListenerForMethod(target, "handlePrint");
        item.addActionListener(listener);
        item.setActionCommand(resBundle.getString("printItemText"));
        item.setEnabled(true);
        chartMenu.add(item, 5);

        //  Add Page Setup to the chart menu.
        item = new JMenuItem(resBundle.getString("pageSetupItemText"));
        listener = AppUtilities.getActionListenerForMethod(target, "handlePageSetup");
        item.addActionListener(listener);
        item.setActionCommand(resBundle.getString("pageSetupItemText"));
        item.setEnabled(true);
        chartMenu.add(item, 5);

        //  Replace the default Save As... item with our own implementation.
        chartMenu.remove(3);

        //  Add our own Save As PNG... item to chart menu.
        item = new JMenuItem(resBundle.getString("saveAsPNGItemText"));
        listener = AppUtilities.getActionListenerForMethod(this, "handleSaveAsPNG");
        item.addActionListener(listener);
        item.setActionCommand(resBundle.getString("saveAsPNGItemText"));
        item.setEnabled(true);
        chartMenu.add(item, 3);

        if (createSVGGraphics2D(10, 10) != null) {
            //  Add our own Save As SVG... item to chart menu.
            item = new JMenuItem(resBundle.getString("saveAsSVGItemText"));
            listener = AppUtilities.getActionListenerForMethod(this, "handleSaveAsSVG");
            item.addActionListener(listener);
            item.setActionCommand(resBundle.getString("saveAsSVGItemText"));
            item.setEnabled(true);
            chartMenu.add(item, 4);
        }
    }

    /**
     * Handles the user choosing "New..." from the File menu. Creates a new application
     * window instance.
     *
     * @param event The event that caused this method to be called.
     */
    public void handleNew(ActionEvent event) {
        MDIApplication.getInstance().handleNew(event);
    }

    /**
     * Handle the user choosing "Close" from the File menu. This implementation dispatches
     * a "Window Closing" event to this window.
     *
     * @param event The event that caused this method to be called.
     */
    public void handleClose(ActionEvent event) {
        this.dispatchEvent(new WindowEvent(this, WindowEvent.WINDOW_CLOSING));
    }

    /**
     * Handle the user choosing "Save As PNG..." from the File menu. Saves the currently
     * displayed plot as a PNG format image file.
     *
     * @param event The event that caused this method to be called. Ignored.
     */
    public void handleSaveAsPNG(ActionEvent event) {
        try {
            //  Go off and actually do the Save As PNG....  Catch exceptions here.
            doSaveAsPNG();

        } catch (IOException e) {
            Logger.getLogger(AbstractPlotWindow.class.getName()).log(Level.SEVERE, null, e);
            String msg = e.getMessage();
            if (msg == null) {
                Throwable cause = e.getCause();
                if (cause != null)
                    msg = cause.getMessage();
            }
            ResourceBundle resBundle = MDIApplication.getInstance().getResourceBundle();
            JOptionPane.showMessageDialog(this, msg, resBundle.getString("ioErrorTitle"), JOptionPane.ERROR_MESSAGE);

        } catch (Exception e) {
            Logger.getLogger(AbstractPlotWindow.class.getName()).log(Level.SEVERE, null, e);
            ResourceBundle resBundle = MDIApplication.getInstance().getResourceBundle();
            AppUtilities.showException(this, resBundle.getString("unexpectedTitle"), resBundle.getString("unexpectedMsg"), e);
        }

    }

    /**
     * Handle the user choosing "Save As SVG..." from the File menu. Saves the currently
     * displayed plot as an SVG format file.
     *
     * @param event The action event that caused this method to be called. Ignored.
     */
    public void handleSaveAsSVG(ActionEvent event) {
        try {
            //  Go off and actually do the Save As SVG....  Catch exceptions here.
            doSaveAsSVG();

        } catch (IOException e) {
            Logger.getLogger(AbstractPlotWindow.class.getName()).log(Level.SEVERE, null, e);
            String msg = e.getMessage();
            if (msg == null) {
                Throwable cause = e.getCause();
                if (cause != null)
                    msg = cause.getMessage();
            }
            ResourceBundle resBundle = MDIApplication.getInstance().getResourceBundle();
            JOptionPane.showMessageDialog(this, msg, resBundle.getString("ioErrorTitle"), JOptionPane.ERROR_MESSAGE);

        } catch (Exception e) {
            Logger.getLogger(AbstractPlotWindow.class.getName()).log(Level.SEVERE, null, e);
            ResourceBundle resBundle = MDIApplication.getInstance().getResourceBundle();
            AppUtilities.showException(this, resBundle.getString("unexpectedTitle"), resBundle.getString("unexpectedMsg"), e);
        }

    }

    /**
     * Handle the user choosing "Page Setup..." from the File menu. Displays a Page Setup
     * dialog allowing the user to change the page settings.
     *
     * @param event The action event that caused this method to be called. Ignored.
     */
    public void handlePageSetup(ActionEvent event) {
        PrinterJob job = PrinterJob.getPrinterJob();
        if (_pf == null)
            _pf = job.defaultPage();
        _pf = job.pageDialog(_pf);
    }

    /**
     * Handle the user choosing "Print" from the File menu. Prints the currently
     * displayed plot.
     *
     * @param event The action event that caused this method to be called.
     * Ignored.
     */
    public void handlePrint(ActionEvent event) {
        PrinterJob job = PrinterJob.getPrinterJob();

        if (_pf == null)
            _pf = job.defaultPage();
        job.setPrintable(getChartPanel(), _pf);

        if (job.printDialog()) {
            try {

                job.print();

            } catch (PrinterException e) {
                Logger.getLogger(AbstractPlotWindow.class.getName()).log(Level.SEVERE, null, e);
                JOptionPane.showMessageDialog(this, e);
            }
        }

    }

    /**
     * Handle the user choosing "Save As PNG..." from the File menu. This asks the user
     * for input and then saves the file. This method throws all exceptions.
     *
     * @return true if the user cancels, false if the file was saved.
     */
    private boolean doSaveAsPNG() throws IOException {
        ResourceBundle resBundle = MDIApplication.getInstance().getResourceBundle();

        //  Build the filename to prompt the user with.
        String extension = "png";
        String dir = MDIApplication.getInstance().getPreferences().getLastPath();
        String fileName = _dataName;
        fileName += "." + extension;

        ExtFilenameFilter fnFilter = new ExtFilenameFilter();
        fnFilter.addExtension(extension);

        //  Ask the user to select a file for saving.
        File theFile = AppUtilities.selectFile4Save(this,
                MessageFormat.format(resBundle.getString("fileSaveDialogMsg"), extension.toUpperCase()),
                dir, fileName, fnFilter, extension,
                resBundle.getString("fileExistsMsgFmt"), resBundle.getString("warningTitle"));

        if (theFile != null) {
            if (theFile.exists() && !theFile.canWrite())
                throw new IOException(resBundle.getString("canNotWrite2File"));

            //  Write out the PNG file.
            ChartPanel plotPanel = getChartPanel();
            ChartUtilities.saveChartAsPNG(theFile, plotPanel.getChart(),
                    plotPanel.getWidth(), plotPanel.getHeight());

            return false;
        }

        return true;
    }

    /**
     * Handle the user choosing "Save As SVG..." from the File menu. This asks the user
     * for input and then saves the file. This method throws all exceptions.
     *
     * @return true if the user cancels, false if the file was saved.
     */
    private boolean doSaveAsSVG() throws IOException {
        ResourceBundle resBundle = MDIApplication.getInstance().getResourceBundle();

        //  Build the filename to prompt the user with.
        String extension = "svg";
        String dir = MDIApplication.getInstance().getPreferences().getLastPath();
        String fileName = _dataName;
        fileName += "." + extension;

        ExtFilenameFilter fnFilter = new ExtFilenameFilter();
        fnFilter.addExtension(extension);

        //  Ask the user to select a file for saving.
        File theFile = AppUtilities.selectFile4Save(this,
                MessageFormat.format(resBundle.getString("fileSaveDialogMsg"), extension.toUpperCase()),
                dir, fileName, fnFilter, extension,
                resBundle.getString("fileExistsMsgFmt"), resBundle.getString("warningTitle"));

        if (theFile != null) {
            if (theFile.exists() && !theFile.canWrite())
                throw new IOException(resBundle.getString("canNotWrite2File"));

            //  Write out the SVG file.
            try (FileOutputStream outputStream = new FileOutputStream(theFile)) {
                ChartPanel plotPanel = getChartPanel();
                saveChartAsSVG(plotPanel.getChart(), outputStream, plotPanel.getWidth(), plotPanel.getHeight());
            }

            return false;
        }

        return true;
    }

    /**
     * Write .svg file based on provided JFreeChart.
     *
     * @param aChart       Chart to be written out as SVG.
     * @param outputStream An OutputStream to a file that will hold the SVG XML.
     * @param aWidth       Width of image.
     * @param aHeight      Height of image.
     * @throws java.io.IOException if there is any problem writing out the SVG file.
     */
    public void saveChartAsSVG(JFreeChart aChart, OutputStream outputStream, int aWidth, int aHeight) throws IOException {

        // Use reflection to get the SVG string
        String svg = generateSVG(aChart, getWidth(), getHeight());
        if (svg == null)
            throw new IOException(MDIApplication.getInstance().getResourceBundle().getString("svgErrMsg"));

        PrintWriter output = new PrintWriter(new OutputStreamWriter(outputStream, "UTF-8"));
        output.println("<!DOCTYPE svg PUBLIC \"-//W3C//DTD SVG 1.1//EN\" \"http://www.w3.org/Graphics/SVG/1.1/DTD/svg11.dtd\">");
        output.println(svg);
        output.flush();

    }

    /**
     * Generates a string containing a rendering of the chart in SVG format. This feature
     * is only supported if the JFreeSVG library is included on the classpath.
     *
     * @return A string containing an SVG element for the current chart, or
     * <code>null</code> if there is a problem with the method invocation by reflection.
     */
    private static String generateSVG(JFreeChart chart, int width, int height) {
        Graphics2D g2 = createSVGGraphics2D(width, height);
        if (g2 == null)
            throw new IllegalStateException("The JFreeSVG library could not be found.");
        
        // We suppress shadow generation, because SVG is a vector format and
        // the shadow effect is applied via bitmap effects...
        g2.setRenderingHint(JFreeChart.KEY_SUPPRESS_SHADOW_GENERATION, true);
        String svg = null;
        Rectangle2D drawArea = new Rectangle2D.Double(0, 0, width, height);
        chart.draw(g2, drawArea);
        try {
            Method m = g2.getClass().getMethod("getSVGElement");
            svg = (String)m.invoke(g2);
        } catch (NoSuchMethodException | SecurityException | IllegalAccessException | IllegalArgumentException | InvocationTargetException ignore) {
            // null will be returned
        }


        return svg;
    }

    private static Graphics2D createSVGGraphics2D(int w, int h) {
        try {

            Class svgGraphics2d = Class.forName("org.jfree.graphics2d.svg.SVGGraphics2D");
            Constructor ctor = svgGraphics2d.getConstructor(int.class, int.class);
            return (Graphics2D) ctor.newInstance(w, h);

        } catch (ClassNotFoundException | NoSuchMethodException | SecurityException | InstantiationException
                | IllegalAccessException | IllegalArgumentException | InvocationTargetException ignore) {
            return null;
        }
    }

    /**
     * Handle the user choosing "Copy" from the Edit menu. Copies the currently displayed
     * plot as an image to the clipboard.
     *
     * @param event The event that caused this method to be called. Ignored.
     */
    public void handleCopy(ActionEvent event) {

        //  Get a copy of the current plot as an image.
        ChartPanel plotPanel = getChartPanel();
        int width = plotPanel.getWidth();
        int height = plotPanel.getHeight();
        JFreeChart chart = plotPanel.getChart();
        Image chartImage = chart.createBufferedImage(width * 2, height * 2, width, height, null);

        //  Place the image in the clipboard.
        setClipboard(chartImage);
    }

    /**
     * Method that places an image on the clipboard.
     */
    private static void setClipboard(Image image) {
        ImageSelection imgSel = new ImageSelection(image);
        Toolkit.getDefaultToolkit().getSystemClipboard().setContents(imgSel, null);
    }
}
