/*
 *  LookupWindow -- Window that allows the user to do table lookups.
 *
 *  Copyright (C) 2000-2024, Joseph A. Huwaldt.
 *  All rights reserved.
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 2
 *  of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *  Or visit:  http://www.gnu.org/licenses/gpl.html
 */

import jahuwaldt.js.tables.Float64Table;
import jahuwaldt.swing.*;
import java.awt.*;
import java.awt.event.*;
import java.text.DecimalFormat;
import static java.text.MessageFormat.format;
import java.text.NumberFormat;
import java.util.ArrayList;
import java.util.Hashtable;
import java.util.List;
import java.util.ResourceBundle;
import java.util.logging.Logger;
import javax.swing.*;
import javax.swing.border.TitledBorder;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;
import javolution.text.Text;
import javolution.util.FastTable;
import org.jscience.mathematics.number.Float64;
import org.jscience.mathematics.vector.Float64Vector;

/**
 * An application window for the application TableReader that displays a set of sliders
 * and text fields for doing table lookups on a selected table.
 *
 * <p> Modified by: Joseph A. Huwaldt </p>
 *
 * @author Joseph A. Huwaldt Date: February 24, 2000
 * @version January 15, 2024
 */
public class LookupWindow extends JFrame {

    //  The logger for this class.
    private static final Logger LOGGER = Logger.getLogger(LookupWindow.class.getName());
    
    // The resolution of the slider scale.
    private static final int SLIDER_SCALE = (Integer.MAX_VALUE) / 100;

    // The tick mark increment for the slider scale.
    private static final int TICK_SPACING = SLIDER_SCALE / 10;

    // A copy of the table we are doing a lookup for.
    private final Float64Table table;

    // The minimum and maximum breakpoints for the table.
    private final double[] minBP;
    private final double[] maxBP;

    // The independent variables we want to do a look up for.
    private final List<Float64> _indep;

    // A copy of the sliders used in this window, one for each dimension.
    private final JSlider[] sliders;

    // A copy of the text fields used by this window, one for each dimension.
    private final DecimalField[] textFields;

    // The text field that holds the result of the lookup.
    private final DecimalField result;

    /**
     * Creates a new instance of a table lookup window using the specified
     * table.
     *
     * @param theTable The table we want to do a table lookup on.
     * @throws java.lang.NoSuchMethodException if there is a problem creating
     * this window.
     */
    @SuppressWarnings({"LeakingThisInConstructor", "OverridableMethodCallInConstructor"})
    public LookupWindow(Float64Table theTable) throws NoSuchMethodException {
        super(format(MDIApplication.getInstance().getResourceBundle().getString("lookupTitle"), theTable.getName()));

        //  Have the window dispose of itself when it closes.
        this.setDefaultCloseOperation(WindowConstants.DISPOSE_ON_CLOSE);

        this.setResizable(true);

        //  Define the window icon in the taskbar.
        ImageIcon icon = AppUtilities.getResourceIcon(MDIApplication.getInstance().getResourceBundle().getString("applicationIconResName"), LOGGER);
        if (icon != null)
            this.setIconImage(icon.getImage());

        // Set up some class variables.
        this.table = theTable;
        int numDims = theTable.getDimensions();
        minBP = new double[numDims];
        maxBP = new double[numDims];

        _indep = FastTable.newInstance();
        for (int i = 0; i < numDims; ++i) {
            _indep.add(Float64.ZERO);
        }

        sliders = new JSlider[numDims];
        textFields = new DecimalField[numDims];

        // Create some event listeners.
        SliderListener sliderL = new SliderListener();
        TextFieldListener textL = new TextFieldListener();
        TextFieldFocusL focusL = new TextFieldFocusL();

        // Define the format we are going to use for the text fields.
        DecimalFormat format = (DecimalFormat)NumberFormat.getNumberInstance();
        format.setMaximumFractionDigits(5);
        format.setMinimumFractionDigits(1);

        // Set up this window's user interface.
        JPanel grid1 = new JPanel(new GridLayout(1, numDims, 10, 10));
        grid1.setBorder(BorderFactory.createEmptyBorder(10, 10, 10, 10));
        grid1.setAlignmentX(JPanel.CENTER_ALIGNMENT);

        // Create a label, slider, and text edit box for each independent variable.
        List<Text> indepNames = theTable.getIndepNames();
        for (int i = 0; i < numDims; ++i) {

            // Create a slider for this independent variable.
            Float64Vector bpoints = theTable.getBreakpoints(i);
            minBP[i] = bpoints.getValue(0);
            maxBP[i] = bpoints.getValue(bpoints.getDimension() - 1);
            _indep.set(i, Float64.valueOf(minBP[i]));			//	_indep[i] = minBP[i];
            JSlider slider = new JSlider(JSlider.VERTICAL, 0, SLIDER_SCALE, 0);
            slider.addChangeListener(sliderL);
            slider.setMajorTickSpacing(TICK_SPACING);
            slider.setPaintTicks(true);
            slider.putClientProperty("dimension", i);

            //Create labels for the top and bottom of the slider.
            @SuppressWarnings("UseOfObsoleteCollectionType")
            Hashtable labelTable = new Hashtable();
            labelTable.put(0, new JLabel(Double.toString(minBP[i])));
            labelTable.put(SLIDER_SCALE,
                    new JLabel(Double.toString(maxBP[i])));
            slider.setLabelTable(labelTable);
            slider.setPaintLabels(true);
            slider.setBorder(BorderFactory.createEmptyBorder(0, 10, 5, 0));
            sliders[i] = slider;

            // Create a text area for this slider.
            DecimalField textField = new DecimalField(minBP[i], 7, format);
            textField.addActionListener(textL);
            textField.addFocusListener(focusL);
            textField.putClientProperty("dimension", i);
            textFields[i] = textField;

            // Layout the independent panel.
            JPanel aPanel = new JPanel(new BorderLayout());
            TitledBorder tborder = BorderFactory.createTitledBorder(indepNames.get(i).toString());
            tborder.setTitleJustification(TitledBorder.CENTER);
            aPanel.setBorder(tborder);
            aPanel.add(slider, BorderLayout.CENTER);
            aPanel.add(textField, BorderLayout.SOUTH);

            // Add the independent panel to the window's grid layout.
            grid1.add(aPanel);
        }

        // Make it so you can tab from on text field to another.
        List<Component> order = new ArrayList<>();
        for (int i = 0; i < numDims; ++i) {
            order.add(textFields[i]);
        }
        this.setFocusTraversalPolicy(new OrderedFocusTraversalPolicy(order));

        // Create a table label.
        JLabel tableL = new JLabel(theTable.getName() + " = ");

        // Create a text field that contains the look up value.
        result = new DecimalField(0, 10, format);
        result.setEditable(false);
        doLookup();

        JPanel botPanel = new JPanel(new FlowLayout());
        botPanel.add(tableL);
        botPanel.add(result);

        // Add everything to the window's pane.
        getContentPane().add(grid1, BorderLayout.CENTER);
        getContentPane().add(botPanel, BorderLayout.SOUTH);

        //  Create a menu bar for this window.
        JMenuBar menuBar = createMenuBar();
        this.setJMenuBar(menuBar);

        //  Size the window.
        this.pack();

        //  Position the window so that each new window can be seen.
        AppUtilities.positionWindow(this, getWidth(), getHeight());
    }
    
    /**
     * A FocusTraversalPolicy that cycles through a supplied list of Component
     * objects.
     */
    public static class OrderedFocusTraversalPolicy
            extends FocusTraversalPolicy {
        private final List<Component> order;
        
        public OrderedFocusTraversalPolicy(List<Component> order) {
            this.order = new ArrayList<>(order.size());
            this.order.addAll(order);
        }

        @Override
        public Component getComponentAfter(Container focusCycleRoot, Component aComponent) {
            int idx = (order.indexOf(aComponent) + 1) % order.size();
            return order.get(idx);
        }
        
        @Override
        public Component getComponentBefore(Container focusCycleRoot, Component aComponent) {
            int idx = order.indexOf(aComponent) - 1;
            if (idx < 0) {
                idx = order.size() - 1;
            }
            return order.get(idx);
        }
        
        @Override
        public Component getDefaultComponent(Container focusCycleRoot) {
            return order.get(0);
        }
        
        @Override
        public Component getLastComponent(Container focusCycleRoot) {
            return order.get(order.size() - 1);
        }
        
        @Override
        public Component getFirstComponent(Container focusCycleRoot) {
            return order.get(0);
        }
    }

    /**
     * Initializes and displays the menus associated with this application.
     */
    private JMenuBar createMenuBar() throws NoSuchMethodException {
        MDIApplication guiApp = MDIApplication.getInstance();
        ResourceBundle resBundle = guiApp.getResourceBundle();

        //  Create a menu bar.
        JMenuBar menuBar = new JMenuBar();

        // Set up the file menu.
        int row = 0;
        String[][] menuStrings = new String[6][3];
        menuStrings[row][0] = resBundle.getString("newItemText");
        menuStrings[row][1] = resBundle.getString("newItemKey");
        menuStrings[row][2] = "handleNew";
        ++row;
        menuStrings[row][0] = resBundle.getString("openItemText");
        menuStrings[row][1] = resBundle.getString("openItemKey");
        menuStrings[row][2] = null;
        ++row;
        ++row;	//	Blank line
        menuStrings[row][0] = resBundle.getString("closeItemText");
        menuStrings[row][1] = resBundle.getString("closeItemKey");
        menuStrings[row][2] = "handleClose";
        ++row;
        menuStrings[row][0] = resBundle.getString("saveItemText");
        menuStrings[row][1] = resBundle.getString("saveItemKey");
        menuStrings[row][2] = null;
        ++row;
        menuStrings[row][0] = resBundle.getString("saveAsItemText");
        menuStrings[row][1] = null;
        menuStrings[row][2] = null;

        JMenu menu = AppUtilities.buildMenu(this, resBundle.getString("fileMenuText"), menuStrings);
        menuBar.add(menu);

        //  Add a Quit menu item (if it isn't automatically present).
        if (!QuitJMenuItem.isAutomaticallyPresent()) {
            menu.addSeparator();
            menu.add(guiApp.getQuitJMenuItem());
        }


        // Set up the edit menu.
        row = 0;
        menuStrings = new String[6][3];
        menuStrings[row][0] = resBundle.getString("undoItemText");
        menuStrings[row][1] = resBundle.getString("undoItemKey");
        menuStrings[row][2] = null;
        ++row;
        menuStrings[row][0] = resBundle.getString("redoItemText");
        menuStrings[row][1] = null;
        menuStrings[row][2] = null;
        ++row;
        ++row;	//	Blank line.
        menuStrings[row][0] = resBundle.getString("cutItemText");
        menuStrings[row][1] = resBundle.getString("cutItemKey");
        menuStrings[row][2] = null;
        ++row;
        menuStrings[row][0] = resBundle.getString("copyItemText");
        menuStrings[row][1] = resBundle.getString("copyItemKey");
        menuStrings[row][2] = null;
        ++row;
        menuStrings[row][0] = resBundle.getString("pasteItemText");
        menuStrings[row][1] = resBundle.getString("pasteItemKey");
        menuStrings[row][2] = null;

        menu = AppUtilities.buildMenu(this, resBundle.getString("editMenuText"), menuStrings);
        menuBar.add(menu);

        //  Create a Window's menu.
        menu = MDIApplication.newWindowsMenu(resBundle.getString("windowsMenuText"));
        menuBar.add(menu);

        //  Create an about menu item (if not automatically present).
        if (!AboutJMenuItem.isAutomaticallyPresent()) {
            //  Create Help menu.
            menu = new JMenu(resBundle.getString("helpMenuText"));
            menuBar.add(menu);

            //  Add the "About" item to the Help menu.
            menu.add(guiApp.getAboutJMenuItem());
        }

        return menuBar;
    }

    /**
     * Handles the user choosing "New..." from the File menu. Creates a new
     * application window instance.
     *
     * @param event The action event that caused this method to be called.
     */
    public void handleNew(ActionEvent event) {
        MDIApplication.getInstance().handleNew(event);
    }

    /**
     * Handle the user choosing "Close" from the File menu. This implementation
     * dispatches a "Window Closing" event to this window.
     *
     * @param event The action event that caused this method to be called.
     */
    public void handleClose(ActionEvent event) {
        this.dispatchEvent(new WindowEvent(this, WindowEvent.WINDOW_CLOSING));
    }

    /**
     * This method actually does the table lookup and stores the result in the
     * result text field.
     */
    private void doLookup() {
        double value = table.lookupValue(_indep);
        result.setValue(value);
    }

    /**
     * Handles the user changing the value in a text field.
     *
     * @param input The text field that has had it's value changed.
     */
    private void doTextFieldChange(DecimalField input) {

        // Find out what dimension this is and what value it has been set to.
        int i = ((Integer)input.getClientProperty("dimension"));
        double value = input.getValue();

        // Validate the user's input.
        if (value > maxBP[i])
            value = maxBP[i];
        else {
            if (value < minBP[i])
                value = minBP[i];
        }

        // Update the slider to match the text input.
        int val = (int)((value - minBP[i]) / (maxBP[i] - minBP[i]) * SLIDER_SCALE);
        sliders[i].setValue(val);

        // Do the table lookup.
        _indep.set(i, Float64.valueOf(value));      // _indep[i] = value;
        doLookup();
    }

    /**
     * Defines a listener for our sliders that is called whenever the slider is
     * changed.
     */
    private class SliderListener implements ChangeListener {

        @Override
        public void stateChanged(ChangeEvent e) {
            JSlider input = (JSlider)e.getSource();

            // Find out what dimension this is and what value it has been set to.
            int i = ((Integer)input.getClientProperty("dimension"));
            int val = input.getValue();

            // Update the text field to match the slider.
            double value = (maxBP[i] - minBP[i]) * val / SLIDER_SCALE + minBP[i];
            textFields[i].setValue(value);

            // Do the table lookup.
            _indep.set(i, Float64.valueOf(value));      // _indep[i] = value;
            doLookup();
        }
    }

    /**
     * Defines a listener for our text fields that is called whenever the user
     * types in a new number and presses enter.
     */
    private class TextFieldListener implements ActionListener {

        @Override
        public void actionPerformed(ActionEvent e) {
            DecimalField input = (DecimalField)e.getSource();
            doTextFieldChange(input);
        }
    }

    /**
     * Defines a listener for our text fields that is called whenever the focus
     * changes off a particular text field.
     */
    private class TextFieldFocusL extends FocusAdapter {

        @Override
        public void focusLost(FocusEvent e) {
            DecimalField input = (DecimalField)e.getSource();
            doTextFieldChange(input);
        }
    }
}
