/*
 *  TablePlotWindow -- Window that plots selected n-dimensional tables.
 *
 *  Copyright (C) 2000-2024, Joseph A. Huwaldt
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 2
 *  of the License, or (at your option) any later version.
 *  
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *  Or visit:  http://www.gnu.org/licenses/gpl.html
 */

import jahuwaldt.js.tables.Float64Table;
import jahuwaldt.swing.AppUtilities;
import jahuwaldt.swing.DecimalField;
import jahuwaldt.swing.MDIApplication;
import jahuwaldt.tools.math.MathTools;
import jahuwaldt.util.GeneralFormat;
import java.awt.BorderLayout;
import java.awt.GridLayout;
import java.awt.SystemColor;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.FocusAdapter;
import java.awt.event.FocusEvent;
import java.text.DecimalFormat;
import java.text.NumberFormat;
import java.util.List;
import java.util.ResourceBundle;
import javax.swing.*;
import javolution.util.FastTable;
import org.jfree.chart.ChartFactory;
import org.jfree.chart.ChartPanel;
import org.jfree.chart.JFreeChart;
import org.jfree.chart.labels.StandardXYToolTipGenerator;
import org.jfree.chart.plot.PlotOrientation;
import org.jfree.chart.plot.XYPlot;
import org.jfree.chart.renderer.xy.XYLineAndShapeRenderer;
import org.jfree.chart.title.LegendTitle;
import org.jfree.data.xy.XYSeries;
import org.jfree.data.xy.XYSeriesCollection;
import org.jscience.mathematics.number.Integer64;
import org.jscience.mathematics.vector.Float64Vector;

/**
 * An application window for the application TableReader that displays a plot of
 * a selected n-dimensional table.
 *
 * <p> Modified by: Joseph A. Huwaldt </p>
 *
 * @author Joseph A. Huwaldt, Date: December 21, 2000
 * @version January 14, 2024
 */
public class TablePlotWindow extends AbstractPlotWindow {

    //  Numerical tolerance.
    private static final double EPS = MathTools.EPS;
    /**
     * Creates the plot of the table showing the 1st & 2nd independents at the
     * current breakpoint for higher independents.
     *
     * @param table The table being plotted.
     * @param nf Number format used to format plot labels.
     * @param index Pre-allocated storage space for indices into the plot table.
     * @param dimOrder List of integers that give the order that the dimensions
     * are to be plotted.
     * @param showCubicSpline A flag indicating if a cubic spline should be
     * drawn through the points of a 1D table.
     * @param undefined The value to use to represent an undefined number that
     * shouldn't be plotted.
     */
    private static JFreeChart createChart(Float64Table table, NumberFormat nf,
            List<Integer64> index, int[] dimOrder, boolean showCubicSpline, double undefined) {

        // List of indexes into the n-dimensional table.
        int dim = table.getDimensions();

        String indepName = table.getIndepName(dimOrder[dim - 1]).toString();

        // Build up a run list.
        Float64Vector xArr = table.getBreakpoints(dimOrder[dim - 1]);
        XYSeriesCollection runList = new XYSeriesCollection();

        if (dim < 2) {
            // Just a single run for 1D plots.
            runList.addSeries(createRun(table, indepName, xArr, index, dimOrder, undefined));

            // If the cubic spline option is enabled, then add a cubic spline through the points.
            if (showCubicSpline) {
                XYSeries series = createCubicSplineRun(table, xArr, undefined);
                if (series != null)
                    runList.addSeries(series);
            }

        } else {
            // All higher dimension plots are represented by a plot with multiple runs and panes.
            String indepName2 = table.getIndepName(dimOrder[dim - 2]).toString();
            Float64Vector d2BrkPoints = table.getBreakpoints(dimOrder[dim - 2]);
            int length = d2BrkPoints.getDimension();
            for (int i = 0; i < length; ++i) {
                index.set(dimOrder[dim - 2], Integer64.valueOf(i));			//	index[dimOrder[dim - 2]] = i;
                String runName = indepName2 + " = " + nf.format(d2BrkPoints.getValue(i));
                runList.addSeries(createRun(table, runName, xArr, index, dimOrder, undefined));
            }
        }

        // Create the chart.
        JFreeChart chart = ChartFactory.createXYLineChart(
                null, // No Title
                indepName, // Domain axis label
                table.getName().toString(), // Range axis label
                runList, // Data to be plotted
                PlotOrientation.VERTICAL, // Horizontal orientation
                true, // Include legend
                true, // Include tooltips
                false);                     // Do not include URLs

        // Move legend to the right hand side of the plot.
        LegendTitle legend = chart.getLegend();
        legend.setPosition(org.jfree.ui.RectangleEdge.RIGHT);

        // Adjust the renderer features.
        XYPlot plot = (XYPlot)chart.getPlot();
        XYLineAndShapeRenderer renderer = new XYLineAndShapeRenderer(true, true);
        renderer.setBaseToolTipGenerator(new StandardXYToolTipGenerator());

        // Correct the line type for the cubic spline if it is used.
        if (dim < 2 && showCubicSpline)
            renderer.setSeriesShapesVisible(1, false);

        plot.setRenderer(renderer);

        return chart;
    }
    /**
     * Creates a single run from a table at the specified dimension.
     */
    private static XYSeries createRun(Float64Table table, String runName, Float64Vector xArr, List<Integer64> index, int[] dimOrder, double undefined) {

        XYSeries run = new XYSeries(runName);
        int dim = table.getDimensions();
        int length = xArr.getDimension();
        for (int i = 0; i < length; ++i) {
            index.set(dimOrder[dim - 1], Integer64.valueOf(i)); // index[dimOrder[dim - 1]] = i;
            double depValue = table.getValue(index);
            if (Math.abs(depValue - undefined) < EPS)
                depValue = Double.NaN;
            run.add(xArr.getValue(i), depValue);
        }

        return run;
    }
    /**
     * Create a single cubic spline run from a 1D table.
     */
    private static XYSeries createCubicSplineRun(Float64Table table, Float64Vector xArr, double undefined) {

        // Get the dependent data.
        Float64Vector yArr = table.getAll().vectorization();
        int nx = xArr.getDimension();
        int nxm1 = nx - 1;

        // Copy the X and Y values to temporary arrays for finding the cubic spline coefficients.
        double[] tau = new double[nx];
        double[][] coefs = new double[4][nx];
        for (int i = 0; i < nx; ++i) {
            tau[i] = xArr.getValue(i);
            double yValue = yArr.getValue(i);
            if (Math.abs(yValue - undefined) < EPS)
                return null;
            coefs[0][i] = yValue;
        }

        // Set the slopes for the ends of the spline.
        coefs[1][0] = 0;
        double denom = tau[1] - tau[0];
        if (Math.abs(denom) > 1e-10)
            coefs[1][0] = (coefs[0][1] - coefs[0][0]) / denom;
        coefs[1][nxm1] = 0;
        denom = tau[nxm1] - tau[nxm1 - 1];
        if (Math.abs(denom) > 1e-10)
            coefs[1][nxm1] = (coefs[0][nxm1] - coefs[0][nxm1 - 1]) / denom;

        // Compute the cubic spline coefficients.
        cubspl(tau, coefs, 1, 1);

        XYSeries run = new XYSeries("Cubic Spline");

        // Create 10 interpolated points in between each original table breakpoint.
        int nPts = 20;
        for (int i = 0; i < nxm1; ++i) {
            double x1 = tau[i];
            double x2 = tau[i + 1];
            double dx = x2 - x1;
            for (int j = 0; j < nPts; ++j) {
                double dx1 = dx / nPts * j;
                double xs = x1 + dx1;
                double depValue = coefs[0][i] + dx1 * (coefs[1][i] + dx1 * (coefs[2][i] + dx1 * coefs[3][i] / 3.) / 2.);
                run.add(xs, depValue);
            }
        }
        run.add(tau[nxm1], coefs[0][nxm1]);

        return run;
    }
    /**
     * Compute the coefficients of a simple cubic spline through a set of X,Y
     * points. From "cubspl.f", "A Practical Guide to Splines", by C. De Boor.
     * Ported from FORTRAN.
     *
     * @param tau An 1D array of abcissae of the data points. Assumed to be
     * strictly increasing.
     * @param coefs A 2D (4 by tau.length) array. Index 0 contains the ordinates
     * of the data points on input. On exit, contains the polynomial
     * coefficients of the cubic interpolating spline with interior knots (or
     * joints) tau(1), ..., tau(n-2). precisely, in the interval (tau(i-1),
     * tau(i)), the spline f is given by: f(x) =
     * c(0,i-1)+h*(c(1,i-1)+h*(c(2,i-1)+h*c(3,i-1)/3.)/2.). where h = x -
     * tau(i-1).
     * @param ibcbeg 0 means no boundary condition at tau(0) is given. In this
     * case, the not-a-knot condition is used, i.e. the jump in the third
     * derivative across tau(1) is forced to zero, thus the first and the second
     * cubic polynomial pieces are made to coincide. 1 means that the slope at
     * tau(0) is made to equal coefs(1,0), supplied by input.
     * @param ibcend 0, 1, or 2 has analogous meaning concerning the boundary
     * condition at tau(n-1), with the additional information taken from
     * coefs(1,n-1).
     */
    private static void cubspl(double[] tau, double[][] coefs, int ibcbeg, int ibcend) {
        // a tridiagonal linear system for the unknown slopes s(i) of
        // f  at tau(i), i=1,...,n, is generated and then solved by gauss elimination,
        // with s(i) ending up in c(2,i), all i.  c(3,.) and c(4,.) are used initially for temporary storage.
        int n = tau.length;
        int nm1 = n - 1;

        // compute first differences of tau sequence and store in c(2,.). also,
        // compute first divided difference of data and store in c(3,.).
        for (int m = 1; m < n; ++m) {
            coefs[2][m] = tau[m] - tau[m - 1];
            coefs[3][m] = (coefs[0][m] - coefs[0][m - 1]) / coefs[2][m];
        }

        // construct first equation from the boundary condition, of the form
        // c(3,0)*s(0) + c(2,0)*s(1) = c(1,0)
        if (ibcbeg < 1) {
            if (n > 2) {
                // not-a-knot condition at left end and n .gt. 2.
                coefs[3][0] = coefs[2][2];
                coefs[2][0] = coefs[2][1] + coefs[2][2];
                coefs[1][0] = ((coefs[2][1] + 2 * coefs[2][0]) * coefs[3][1] * coefs[2][2]
                        + coefs[2][1] * coefs[2][1] * coefs[3][2]) / coefs[2][0];

            } else {
                // no condition at left end and n = 2.
                coefs[3][0] = 1;
                coefs[2][0] = 1;
                coefs[1][0] = 2 * coefs[3][1];
                cubspl400(coefs, ibcbeg, ibcend);
                return;
            }

        } else if (ibcbeg == 1) {
            // slope prescribed at left end.
            coefs[3][0] = 1;
            coefs[2][0] = 0;

        } else {
            // second derivative prescribed at left end.
            coefs[3][0] = 2;
            coefs[2][0] = 1;
            coefs[1][0] = 3. * coefs[3][1] - coefs[2][1] / 2. * coefs[1][0];
        }
        if (n == 2) {
            cubspl400(coefs, ibcbeg, ibcend);
            return;
        }

        // if there are interior knots, generate the corresp. equations and
        // carry out the forward pass of gauss elimination, after which
        // the m-th equation reads    c(3,m)*s(m) + c(2,m)*s(m+1) = c(1,m).
        for (int m = 1; m < nm1; ++m) {
            double g = -coefs[2][m + 1] / coefs[3][m - 1];
            coefs[1][m] = g * coefs[1][m - 1] + 3. * (coefs[2][m] * coefs[3][m + 1] + coefs[2][m + 1] * coefs[3][m]);
            coefs[3][m] = g * coefs[2][m - 1] + 2. * (coefs[2][m] + coefs[2][m + 1]);
        }

        //	construct last equation from the second boundary condition, of the form
        //		(-g*c(3,nm1-1))*s(nm1-1) + c(3,nm1)*s(nm1) = c(1,nm1)
        //	if slope is prescribed at right end, one can go directly to back-
        //	substitution, since c array happens to be set up just right for it
        //	at this point.
        if (ibcend < 1) {
            if (n != 3 || ibcbeg != 0) {
                //	not-a-knot and n .ge. 3, and either n.gt.3 or  also
                //	not-a-knot at left end point.
                double g = coefs[2][nm1 - 1] + coefs[2][nm1];
                coefs[1][nm1] = ((coefs[2][nm1] + 2. * g) * coefs[3][nm1] * coefs[2][nm1 - 1]
                        + coefs[2][nm1] * coefs[2][nm1] * (coefs[0][nm1 - 1] - coefs[0][nm1 - 2]) / coefs[2][nm1 - 1]) / g;
                g = -g / coefs[3][nm1 - 1];
                coefs[3][nm1] = coefs[2][nm1 - 1];
                cubspl500(coefs, g);
                return;
            }
        } else if (ibcend == 1) {
            cubspl600(coefs);
            return;

        }

        cubspl300(coefs);

    }
    private static void cubspl200(double[][] coefs) {
        //	either (n=3 and not-a-knot also at left) or (n=2 and not not-a-knot at left end point).
        int n = coefs[0].length;
        int nm1 = n - 1;

        coefs[1][nm1] = 2 * coefs[3][nm1];
        coefs[3][nm1] = 1;
        double g = -1 / coefs[3][nm1 - 1];

        cubspl500(coefs, g);
    }
    private static void cubspl300(double[][] coefs) {
        //	second derivative prescribed at right endpoint.
        int nm1 = coefs[0].length - 1;

        coefs[1][nm1] = 3. * coefs[3][nm1] + coefs[2][nm1] / 2. * coefs[1][nm1];
        coefs[3][nm1] = 2;
        double g = -1. / coefs[3][nm1 - 1];

        cubspl500(coefs, g);
    }
    private static void cubspl400(double[][] coefs, int ibcbeg, int ibcend) {
        int nm1 = coefs[0].length - 1;
        if (ibcend < 1) {
            if (ibcbeg > 0) {
                cubspl200(coefs);
                return;
            }
            //	not-a-knot at right endpoint and at left endpoint and n = 2.
            coefs[1][nm1] = coefs[3][nm1];
            cubspl600(coefs);

        } else if (ibcend == 1) {
            cubspl600(coefs);

        } else {
            cubspl300(coefs);
        }

    }
    private static void cubspl500(double[][] coefs, double g) {
        int nm1 = coefs[0].length - 1;

        //	complete forward pass of gauss elimination.
        coefs[3][nm1] = g * coefs[2][nm1 - 1] + coefs[3][nm1];
        coefs[1][nm1] = (g * coefs[1][nm1 - 1] + coefs[1][nm1]) / coefs[3][nm1];

        cubspl600(coefs);
    }
    private static void cubspl600(double[][] c) {
        //	carry out back substitution
        int n = c[0].length;
        int l = n - 1;
        int j = l - 1;

        while (true) {
            c[1][j] = (c[1][j] - c[2][j] * c[1][j + 1]) / c[3][j];
            --j;
            if (j <= -1) {
                //	generate cubic coefficients in each interval, i.e., the
                //	deriv.s at its left endpoint, from value and slope at its endpoints.
                for (int i = 1; i < n; ++i) {
                    double dtau = c[2][i];
                    double divdf1 = (c[0][i] - c[0][i - 1]) / dtau;
                    double divdf3 = c[1][i - 1] + c[1][i] - 2. * divdf1;
                    c[2][i - 1] = 2. * (divdf1 - c[1][i - 1] - divdf3) / dtau;
                    c[3][i - 1] = (divdf3 / dtau) * (6. / dtau);
                }
                break;
            }
        }
    }
    /**
     * Create a high dimension breakpoint panel. These are for breakpoints shown
     * as different plot panes.
     */
    private static JPanel createBPPanel(Float64Table table, NumberFormat nf, List<Integer64> index, int[] dimOrder) {
        JPanel thePanel = null;

        int dim = table.getDimensions();
        if (dim > 2) {
            int numElements = dim - 2;
            thePanel = new JPanel(new GridLayout(numElements, 1));

            for (int i = 0; i < numElements; ++i) {
                Float64Vector breakpoints = table.getBreakpoints(dimOrder[i]);
                String indepName = table.getIndepName(dimOrder[i]).toString();
                double bpValue = breakpoints.getValue(index.get(dimOrder[i]).intValue());
                String message = indepName + " = " + nf.format(bpValue);
                JLabel label = new JLabel(message, SwingConstants.CENTER);
                thePanel.add(label);
            }
        }

        return thePanel;
    }

    //	The content pane for our window.
    private final JPanel _centerPanel = new JPanel(new BorderLayout());

    //	References to the left/right buttons.
    private final JButton _leftButton;
    private final JButton _rightButton;

    // A copy of the table we are doing the plot of.
    private final Float64Table _table;

    // A reference to the panel containing the current high dimension breakpoints.
    private JPanel _bpPanel;

    // The number format used to format numbers.
    private final DecimalFormat _nf = (DecimalFormat)NumberFormat.getInstance();

    //	Storage for indexes into multi-dimensioned tables.
    private final List<Integer64> _index;

    //	Indicies that give the order of the plotting of dimensions in the table.
    private final int[] _dimOrder;

    //	Flag indicating if a cubic spline through the table data is shown or not.
    private final JCheckBox _showCubicSplineCB;

    //  A text field for inputting the number that is to be represented as undefined.
    private final DecimalField _undefinedTF;

    /**
     * Creates a new instance of a table plot window using the specified table.
     *
     * @param title The title for the window.
     * @param theTable The table we want to do a plot of.
     * @throws java.lang.NoSuchMethodException if there is any problem creating
     * this window.
     */
    @SuppressWarnings("OverridableMethodCallInConstructor")
    public TablePlotWindow(String title, Float64Table theTable) throws NoSuchMethodException {
        super(title, theTable.getName().toString());

        ResourceBundle resBundle = MDIApplication.getInstance().getResourceBundle();

        setResizable(true);
        setSize(1020, 640);

        //  Position the window so that each new window can be seen.
        AppUtilities.positionWindow(this, getWidth(), getHeight());

        //  Have the window dispose of itself when it closes.
        setDefaultCloseOperation(WindowConstants.DISPOSE_ON_CLOSE);

        // Set up some class variables.
        _table = theTable;
        _nf.setMinimumFractionDigits(3);

        int numDim = theTable.getDimensions();
        _index = FastTable.newInstance();
        for (int i = 0; i < numDim; ++i) {
            _index.add(Integer64.ZERO);
        }

        _dimOrder = new int[numDim];
        for (int i = 0; i < numDim; ++i) {
            _dimOrder[i] = i;
        }

        // Set up this window's user interface.
        getContentPane().add(_centerPanel);

        //  Create an "undefined" value entry field.
        JLabel undefinedLabel = new JLabel(resBundle.getString("undefinedLabel") + " ");
        _undefinedTF = new DecimalField(Double.NaN, 8, new GeneralFormat());
        _undefinedTF.addActionListener(new ActionListener() {
            @Override
            public void actionPerformed(ActionEvent e) {
                regeneratePlot();
            }
        });
        _undefinedTF.addFocusListener(new FocusAdapter() {
            @Override
            public void focusLost(FocusEvent e) {
                regeneratePlot();
            }
        });
        _undefinedTF.setMaximumSize(_undefinedTF.getPreferredSize());
        _undefinedTF.setToolTipText(resBundle.getString("undefinedToolTip"));
        Box undefBox = Box.createHorizontalBox();
        undefBox.add(undefinedLabel);
        undefBox.add(_undefinedTF);

        //	Create navigation buttons.
        JButton fullLeftBtn = new JButton(resBundle.getString("fullLeftBtnText"));
        fullLeftBtn.setToolTipText(resBundle.getString("fullLeftBtnToolTip"));
        fullLeftBtn.setHorizontalAlignment(SwingConstants.CENTER);
        fullLeftBtn.addActionListener(new ActionListener() {
            @Override
            public void actionPerformed(ActionEvent evt) {
                doFullLeftButton();
            }
        });

        _leftButton = new JButton(resBundle.getString("leftBtnText"));
        _leftButton.setToolTipText(resBundle.getString("leftBtnToolTip"));
        _leftButton.setEnabled(false);
        _leftButton.setHorizontalAlignment(SwingConstants.CENTER);
        _leftButton.addActionListener(new ActionListener() {
            @Override
            public void actionPerformed(ActionEvent evt) {
                doLeftButton();
            }
        });

        _rightButton = new JButton(resBundle.getString("rightBtnText"));
        _rightButton.setToolTipText(resBundle.getString("rightBtnToolTip"));
        _rightButton.setEnabled(_table.getDimensions() > 2);
        _rightButton.setHorizontalAlignment(SwingConstants.CENTER);
        _rightButton.addActionListener(new ActionListener() {
            @Override
            public void actionPerformed(ActionEvent evt) {
                doRightButton();
            }
        });

        JButton fullRightBtn = new JButton(resBundle.getString("fullRightBtnText"));
        fullRightBtn.setToolTipText(resBundle.getString("fullRightBtnToolTip"));
        fullRightBtn.setHorizontalAlignment(SwingConstants.CENTER);
        fullRightBtn.addActionListener(new ActionListener() {
            @Override
            public void actionPerformed(ActionEvent evt) {
                doFullRightButton();
            }
        });

        JButton xPlotBtn = new JButton(resBundle.getString("xPlotBtnText"));
        xPlotBtn.setToolTipText(resBundle.getString("xPlotBtnToolTip"));
        xPlotBtn.setHorizontalAlignment(SwingConstants.CENTER);
        xPlotBtn.addActionListener(new ActionListener() {
            @Override
            public void actionPerformed(ActionEvent evt) {
                doXPlotButton();
            }
        });

        JButton closeButton = new JButton(resBundle.getString("closeBtnText"));
        this.getRootPane().setDefaultButton(closeButton);
        closeButton.addActionListener(new ActionListener() {
            @Override
            public void actionPerformed(ActionEvent evt) {
                handleClose(null);
            }
        });

        // Create a button box and add all the buttons to it.
        Box btnBox = new Box(BoxLayout.X_AXIS);
        btnBox.setBackground(SystemColor.window);
        btnBox.add(undefBox);
        btnBox.add(Box.createHorizontalStrut(20));
        btnBox.add(Box.createGlue());

        // Create a cubic spline input
        _showCubicSplineCB = new JCheckBox(resBundle.getString("cubicSplineCBText"), false);
        if (_table.getDimensions() == 1) {
            _showCubicSplineCB.addActionListener(new ActionListener() {
                @Override
                public void actionPerformed(ActionEvent evt) {
                    //	Draw the plot with the new cubic spline option.
                    regeneratePlot();
                }
            });

            btnBox.add(_showCubicSplineCB);
            btnBox.add(Box.createHorizontalStrut(10));
        }

        btnBox.add(fullLeftBtn);
        btnBox.add(_leftButton);
        btnBox.add(_rightButton);
        btnBox.add(fullRightBtn);
        btnBox.add(Box.createHorizontalStrut(20));
        btnBox.add(xPlotBtn);
        btnBox.add(Box.createGlue());
        btnBox.add(Box.createHorizontalStrut(20));
        btnBox.add(closeButton);
        if (_table.getDimensions() < 3) {
            fullLeftBtn.setEnabled(false);
            _leftButton.setEnabled(false);
            _rightButton.setEnabled(false);
            fullRightBtn.setEnabled(false);
        }
        if (_table.getDimensions() < 2) {
            xPlotBtn.setEnabled(false);
        }

        JPanel buttonPanel = new JPanel(new BorderLayout());
        buttonPanel.add(btnBox, BorderLayout.SOUTH);
        buttonPanel.setBorder(BorderFactory.createEmptyBorder(10, 10, 10, 10));

        _centerPanel.add(buttonPanel, BorderLayout.SOUTH);

        // Create the plot panel containing the plot itself.
        ChartPanel plotPanel = createPlotPanel(theTable, _nf, _index, this);
        _centerPanel.add(plotPanel, BorderLayout.CENTER);

        // Create a higher order breakpoint panel if necessary.
        _bpPanel = createBPPanel(_table, _nf, _index, _dimOrder);
        if (_bpPanel != null)
            _centerPanel.add(_bpPanel, BorderLayout.NORTH);

    }

    /**
     * Method that will build an "Export" menu containing a list of all the
     * TableReader objects that are capable of writing to files.
     *
     * @param title	The title for the menu that is created.
     * @return A JMenu instance containing a list of all TableReader objects
     * that can write to files in various formats. If there are no appropriate
     * readers available, null is returned.
     */
    @Override
    protected JMenu buildExportMenu(String title) {
        return null;
    }

    /**
     * Method that regenerates and redraws the plot in the window. The old plot
     * is disposed of and a new plot and plot panel is created.
     */
    private void regeneratePlot() {

        try {
            //	Remove the old plot.
            if (getChartPanel() != null)
                _centerPanel.remove(getChartPanel());

            //	Create the new plot and add it to the window.
            ChartPanel plotPanel = createPlotPanel(_table, _nf, _index, this);
            _centerPanel.add(plotPanel, BorderLayout.CENTER);

            //	Remove the high dimension breakpoint panel.
            if (_bpPanel != null)
                _centerPanel.remove(_bpPanel);

            //	Create a new breakpoint label at the top (if necissary).
            _bpPanel = createBPPanel(_table, _nf, _index, _dimOrder);
            if (_bpPanel != null)
                _centerPanel.add(_bpPanel, BorderLayout.NORTH);

            //	Have the window re-configure itself.
            validate();

        } catch (NoSuchMethodException e) {
            System.err.println("Menu action listener method not found!");
            e.printStackTrace();
        }

    }

    /**
     * Creates a panel that contains the plot of the data.
     *
     * @param theTable The table being plotted.
     * @param nf Number format used to format plot labels.
     * @param index Pre-allocated storage space for indices into the plot table.
     * @param target The object that receives the action events. Must have the
     * following methods defined: "handlePrint(ActionEvent e)", and
     * "handlePageSetup(ActionEvent e)".
     */
    private ChartPanel createPlotPanel(Float64Table theTable, NumberFormat nf,
            List<Integer64> index, Object target) throws NoSuchMethodException {

        // Create the chart.
        JFreeChart chart = createChart(theTable, nf, index, _dimOrder, _showCubicSplineCB.isSelected(), _undefinedTF.getValue());

        // Create a panel to put the chart in.
        createChartPanel(chart, target);

        // Return a reference to the newly created chart panel.
        return getChartPanel();
    }


    /**
     * Handles the user clicking on the full-left button.
     */
    private void doFullLeftButton() {
        _leftButton.setEnabled(false);
        _rightButton.setEnabled(true);

        int dim = _table.getDimensions() - 3;
        for (int i = 0; i <= dim; ++i) {
            _index.set(_dimOrder[i], Integer64.ZERO); // _index[dimOrder[i]] = 0;
        }

        regeneratePlot();
    }

    /**
     * Handles the user clicking on the left button.
     */
    private void doLeftButton() {
        _rightButton.setEnabled(true);
        int dim = _table.getDimensions() - 3;
        int old_dim = dim;

        _index.set(_dimOrder[dim], _index.get(_dimOrder[dim]).minus(Integer64.ONE));    // --_index[dimOrder[dim]];
        while (dim != 0 && _index.get(_dimOrder[dim]).intValue() < 0) {
            int bpIdx = _table.getNumBreakpoints(_dimOrder[dim]) - 1;
            _index.set(_dimOrder[dim], Integer64.valueOf(bpIdx));                       // _index[dimOrder[dim]] = bpIdx;
            --dim;
            _index.set(_dimOrder[dim], _index.get(_dimOrder[dim]).minus(Integer64.ONE)); // --_index[dimOrder[dim]];
        }

        // Have we reached the end of all the breakpoints of
        // all the dimensions of the table?
        dim = old_dim;
        while (dim != 0 && _index.get(_dimOrder[dim]).intValue() == 0) {
            --dim;
        }
        if (dim == 0 && _index.get(_dimOrder[dim]).intValue() == 0)
            _leftButton.setEnabled(false);

        regeneratePlot();
    }

    /**
     * Handles the user clicking on the right button.
     */
    private void doRightButton() {
        _leftButton.setEnabled(true);
        int dim = _table.getDimensions() - 3;
        int old_dim = dim;

        _index.set(_dimOrder[dim], _index.get(_dimOrder[dim]).plus(Integer64.ONE)); // ++_index[dimOrder[dim]];
        while (dim != 0 && _index.get(_dimOrder[dim]).intValue() == _table.getNumBreakpoints(_dimOrder[dim])) {
            _index.set(_dimOrder[dim], Integer64.ZERO); // _index[dimOrder[dim]] = 0;
            --dim;
            _index.set(_dimOrder[dim], _index.get(_dimOrder[dim]).plus(Integer64.ONE)); // ++_index[dimOrder[dim]];
        }

        // Have we reached the end of all the breakpoints of
        // all the dimensions of the table?
        dim = old_dim;
        while (dim != 0 && _index.get(_dimOrder[dim]).intValue() == _table.getNumBreakpoints(_dimOrder[dim]) - 1) {
            --dim;
        }
        if (dim == 0 && _index.get(_dimOrder[dim]).intValue() == _table.getNumBreakpoints(_dimOrder[dim]) - 1)
            _rightButton.setEnabled(false);

        regeneratePlot();
    }

    /**
     * Handles the user clicking on the full-right button.
     */
    private void doFullRightButton() {
        _leftButton.setEnabled(true);
        _rightButton.setEnabled(false);

        int dim = _table.getDimensions() - 3;
        for (int i = 0; i <= dim; ++i) {
            int bpIdx = _table.getNumBreakpoints(_dimOrder[i]) - 1;
            _index.set(_dimOrder[i], Integer64.valueOf(bpIdx)); // _index[dimOrder[i]] = bpIdx;
        }

        regeneratePlot();
    }

    /**
     * Handles the user clicking on the X-Plot button.
     */
    private void doXPlotButton() {

        if (_table.getDimensions() == 2) {
            int tmp = _dimOrder[0];
            _dimOrder[0] = _dimOrder[1];
            _dimOrder[1] = tmp;

        } else {

            //  If there is more than two dimensions, 
            //	ask the user to select the order of the dimensions to plot.
            ResourceBundle resBundle = MDIApplication.getInstance().getResourceBundle();
            SelectPlotDimsDialog dimDialog = new SelectPlotDimsDialog(this, resBundle.getString("selectPlotDimsTitle"),
                    resBundle.getString("selectPlotDimsText"), _table, _dimOrder);
            dimDialog.setLocation(AppUtilities.dialogPosition(dimDialog));
            dimDialog.setVisible(true);

            //	Retrieve the selected cases.
            List<Integer> dimensions = dimDialog.getSelected();
            dimDialog.dispose();
            if (dimensions == null)
                return; // User canceled.

            //	Copy the output from the dialog to the dimensions order array.
            int dims = _dimOrder.length;
            for (int i = 0; i < dims; ++i) {
                Integer element = dimensions.get(i);
                _dimOrder[i] = element;
            }

            //	Reset the plot to the left end of the 3rd independent variable.
            doFullLeftButton();

        }

        //	Draw the plot with the new order of dimensions.
        regeneratePlot();
    }

}
