/*
 *   CSVReader -- A class that can read and write specially organized CSV table file.
 *
 *   Copyright (C) 2000-2024, by Joseph A. Huwaldt.
 *   All rights reserved.
 *   
 *   This library is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU Lesser General Public
 *   License as published by the Free Software Foundation; either
 *   version 2 of the License, or (at your option) any later version.
 *   
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   Library General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *  Or visit:  http://www.gnu.org/licenses/lgpl.html
 */
package jahuwaldt.js.tables;

import jahuwaldt.io.FileUtils;
import static jahuwaldt.js.tables.FTableReader.RESOURCES;
import jahuwaldt.js.util.TextTokenizer;
import jahuwaldt.util.GeneralFormat;
import java.io.*;
import java.text.DateFormat;
import java.text.MessageFormat;
import static java.text.MessageFormat.format;
import java.util.Date;
import java.util.Locale;
import java.util.NoSuchElementException;
import static java.util.Objects.requireNonNull;
import javolution.text.Text;
import javolution.util.FastTable;
import org.jscience.mathematics.number.Float64;
import org.jscience.mathematics.number.Integer64;
import org.jscience.mathematics.vector.Float64Vector;

/**
 * A class that provides methods for reading and writing a specially organized CSV
 * (Comma-Separated-Value) multi-table file.
 * 
 * <p> Modified by: Joseph A. Huwaldt </p>
 * 
* @author Joseph A. Huwaldt, Date: April 3, 2000
 * @version January 15, 2024
 */
public class CSVReader implements FTableReader {

    //  The preferred file extension for files of this reader's type.
    private static final String EXTENSION = "csv";
    
    //  The delimiter used when parsing the file.
    private static final String DELIMITER = ",";

    //  A space.
    private static final Text SPACE = Text.intern(" ");

    /**
     * Returns a string representation of the object. This will return a brief description
     * of the format read by this reader.
     */
    @Override
    public String toString() {
        return RESOURCES.getString("csvDescription");
    }

    /**
     * Returns the preferred file extension (not including the ".") for files of this
     * reader's type.
     * 
     * @return The file extension ("csv").
     */
    @Override
    public String getExtension() {
        return EXTENSION;
    }

    /**
     * Method that determines if this reader can read data from the specified input
     * stream. Will return FTableReader.MAYBE if the file name has the extension "csv".
     *
     * @param name  The name of the file.
     * @param input An input stream containing the data to be read. Any methods that read
     *              from this stream must first set a mark and then reset back to that
     *              mark before the method returns (even if it returns with an exception).
     * @return FTableReader.NO if the file is not recognized at all or FTableReader.MAYBE
     *         if the filename extension is "csv".
     */
    @Override
    public int canReadData(String name, BufferedInputStream input) throws IOException {
        requireNonNull(name, MessageFormat.format(RESOURCES.getString("paramNullErr"), "name"));

        int response = NO;
        String ext = FileUtils.getExtension(name);
        if (EXTENSION.equals(ext))
            response = MAYBE;

        return response;
    }

    /**
     * Returns true. This class can write data to a CSV formatted file.
     */
    @Override
    public boolean canWriteData() {
        return true;
    }

    /**
     * Compares this object with the specified object for order based on the
     * <code>toString().compareTo(o.toString())</code> method. Returns a negative integer,
     * zero, or a positive integer as this object is less than, equal to, or greater than
     * the specified object.
     */
    @Override
    public int compareTo(FTableReader o) {
        return this.toString().compareTo(o.toString());
    }

    /**
     * Method that reads in specially organized CSV formatted data from the specified
     * input stream and returns that data as an {@link FTableList} object.
     *
     * @param input An input stream containing the CSV formatted data.
     * @return An {@link FTableList} object that contains the data read in from the
     *         specified stream.
     * @throws IOException If there is a problem reading the specified stream.
     */
    @Override
    public FTableList read(InputStream input) throws IOException {
        requireNonNull(input, MessageFormat.format(RESOURCES.getString("paramNullErr"), "input"));
        
        //  CSV files are required to be in ASCII with U.S. style number formatting.
        //  Get the default locale and save it.  Then change the default locale to US.
        Locale defLocale = Locale.getDefault();
        Locale.setDefault(Locale.US);

        // Create a reader that can parse the input stream.
        try {
            LineNumberReader reader = new LineNumberReader(new InputStreamReader(input));

            FTableList db = readCSVStream(reader);

            return db;

        } finally {
            //  Restore the default locale.
            Locale.setDefault(defLocale);
        }
    }

    /**
     * Method for writing out all the data stored in the specified list of
     * {@link Float64Table} objects to the specified output stream in CSV format. The CSV
     * format requires that each table have exactly the same set of breakpoints. If the
     * tables do not have the same breakpoints, an exception will be thrown.
     *
     * @param output The output stream to which the data is to be written.
     * @param tables The list of <code>Float64Table</code> objects to be written out.
     * @throws IOException If there is a problem writing to the specified stream.
     */
    @Override
    public void write(OutputStream output, FTableList tables) throws IOException {
        requireNonNull(output, MessageFormat.format(RESOURCES.getString("paramNullErr"), "output"));
        if (tables.isEmpty())
            return;

        // Check for consistant tables.
        checkTables(tables);

        //  CSV files are required to be in ASCII with U.S. style number formatting.
        //  Get the default locale and save it.  Then change the default locale to US.
        Locale defLocale = Locale.getDefault();
        Locale.setDefault(Locale.US);

        // Get a reference to the output stream writer.
        try {
            BufferedWriter writer = new BufferedWriter(new OutputStreamWriter(output));

            // Write out the database notes as table file comments (if there are any).
            writeFileComments(writer, tables);

            // Write out the tables to the output stream.
            writeCSVFile(writer, tables);

            //  Flush the output.
            writer.flush();

        } finally {
            //  Restore the default locale.
            Locale.setDefault(defLocale);
        }
    }

    /**
     * Read in a Comma Separated Value (CSV) formatted table stream and return a new table
     * database containing them.
     *
     * @param reader A reader for reading in the CSV data.
     * @return A collection of tables read in from the input reader.
     */
    private static FTableList readCSVStream(LineNumberReader reader) throws IOException {
        requireNonNull(reader, MessageFormat.format(RESOURCES.getString("paramNullErr"), "reader"));
        
        // Set up the tokenizer options.
        TextTokenizer tokenizer = TextTokenizer.newInstance();
        tokenizer.setDelimiters(DELIMITER);

        // Read in four comment lines.
        FastTable<Text> comments = readComments(reader, tokenizer);

        // Read in the number of breakpoints for each independent variable.
        FastTable<Integer64> indepNum = readNumBreakpoints(reader, tokenizer);
        int numDims = indepNum.size();

        // Read in the names of the independent variables.
        FastTable<Text> indepNames = readIndepNames(reader, tokenizer, numDims);

        // Read in the table names.
        FastTable<Text> tblNames = FastTable.newInstance();
        for (Text token : tokenizer) {
            tblNames.add(token.trim());
        }

        if (tblNames.isEmpty())
            throw new IOException(format(RESOURCES.getString("csvMissingDepNamesMsg"), reader.getLineNumber()));

        // Create empty breakpoints correctly sized for each dimension.
        int numElements = 1;
        FastTable<Float64Vector> breakpoints = FastTable.newInstance();
        for (int i = 0; i < numDims; ++i) {
            int num = indepNum.get(i).intValue();
            FastTable<Float64> row = FastTable.newInstance();
            for (int j = 0; j < num; ++j)
                row.add(Float64.ZERO);
            breakpoints.add(Float64Vector.valueOf(row));
            FastTable.recycle(row);
            numElements *= num;
        }
        FastTable.recycle(indepNum);

        // Create an array of empty tables, with empty breakpoints, one for each dependent variable.
        FTableList tables = FTableList.newInstance();
        for (Text name : tblNames) {
            Float64Table table = Float64Table.valueOf(name, indepNames, breakpoints, null);
            tables.add(table);
        }
        FastTable.recycle(tblNames);

        //  Add any file level notes to the list of tables.
        tables.addAllNotes(comments);
        FastTable.recycle(comments);

        // Read in all the independent and dependent values (one line at a time).
        FastTable<Integer64> pos = FastTable.newInstance();
        FastTable<Float64> old = FastTable.newInstance();
        for (int i = 0; i < numDims; ++i) {
            pos.add(Integer64.ZERO);
            old.add(Float64.ZERO);
        }

        for (int i = 0; i < numElements; ++i) {
            // Read in the next line from the file.
            String aLine = reader.readLine();
            if (aLine == null)
                throw new IOException(format(RESOURCES.getString("EOFErrMsg"), reader.getLineNumber()));
            tokenizer.setText(aLine);

            //  Read in the independent values.
            for (int j = 0; j < numDims; ++j) {
                try {

                    // Read in the independent values.
                    Float64 value = Float64.valueOf(tokenizer.nextToken().trim());
                    if (i == 0) {
                        // First time through, store off old values.
                        old.set(j, value);

                    } else {
                        // Only increment this dimension's index if the value has changed.
                        if (!value.equals(old.get(j))) {
                            old.set(j, value);
                            pos.set(j, pos.get(j).plus(Integer64.ONE));     //  ++pos[j]
                            // Once we've reached the max breakpoint, go back to the 1st.
                            if (pos.get(j).intValue() >= breakpoints.get(j).getDimension())
                                pos.set(j, Integer64.ZERO);
                        }
                    }

                    //  Change the breakpoint value in all the tables.
                    for (Float64Table table : tables)
                        table.setBreakpoint(j, pos.get(j).intValue(), value);

                } catch (NoSuchElementException e) {
                    throw new IOException(format(RESOURCES.getString("csvMissingValueMsg"),
                            indepNames.get(j), reader.getLineNumber()), e);
                    
                } catch (NumberFormatException e) {
                    throw new IOException(format(RESOURCES.getString("csvCantParseValueMsg"), 
                            indepNames.get(j), reader.getLineNumber()), e);
                }
            }

            // Read in the dependent values.
            for (Float64Table table : tables) {
                try {
                    Text token = tokenizer.nextToken();
                    //System.out.println("token = " + token);
                    Float64 value = Float64.valueOf(token.trim());
                    table.set(pos, value);

                } catch (NoSuchElementException e) {
                    throw new IOException(format(RESOURCES.getString("csvMissingValueMsg"), 
                            table.getName(), reader.getLineNumber()), e);
                } catch (NumberFormatException e) {
                    throw new IOException(format(RESOURCES.getString("csvCantParseValueMsg"), 
                            table.getName(), reader.getLineNumber()), e);
                } catch (Exception e) {
                    throw new IOException(format(RESOURCES.getString("csvCantParseValueMsg"), 
                            table.getName(), reader.getLineNumber()), e);
                }
            }
        }

        FastTable.recycle(indepNames);
        FastTable.recycle(breakpoints);
        FastTable.recycle(pos);
        FastTable.recycle(old);
        TextTokenizer.recycle(tokenizer);

        return tables;
    }

    /**
     * Returns any comments read in from the 1st 4 lines of the file.
     *
     * @param reader    A reader for the CSV data.
     * @param tokenizer The tokenizer used to parse the data with the correct delimiter
     *                  set.
     */
    private static FastTable<Text> readComments(LineNumberReader reader, TextTokenizer tokenizer) throws IOException {

        // Read in four comment lines.
        FastTable<Text> comments = FastTable.newInstance();
        for (int i = 0; i < 4; ++i) {
            String aLine = reader.readLine();
            if (aLine == null)
                throw new IOException(format(RESOURCES.getString("EOFErrMsg"), reader.getLineNumber()));

            tokenizer.setText(aLine);
            Text note = Text.EMPTY;
            for (Text token : tokenizer)
                note = note.concat(token.concat(SPACE));

            note = note.trim();
            if (!note.equals(Text.EMPTY))
                comments.add(note);
        }

        return comments;
    }

    /**
     * Return the the number of breakpoints in each dimension of the tables in the file.
     *
     * @param reader    A reader for the CSV data.
     * @param tokenizer The tokenizer used to parse the data with the correct delimiter
     *                  set.
     */
    private static FastTable<Integer64> readNumBreakpoints(LineNumberReader reader, TextTokenizer tokenizer) throws IOException {

        FastTable<Integer64> indepNum = FastTable.newInstance();

        String aLine = reader.readLine();
        if (aLine == null)
            throw new IOException(format(RESOURCES.getString("EOFErrMsg"), reader.getLineNumber()));
        tokenizer.setText(aLine);
        try {
            for (Text token : tokenizer) {
                Integer64 value = Integer64.valueOf(token.trim());
                indepNum.add(value);
            }
        } catch (NumberFormatException e) {
            //  Stop reading if anything other than a number is encountered.
        }

        if (indepNum.isEmpty())
            throw new IOException(format(RESOURCES.getString("csvMissingDimsMsg"), reader.getLineNumber()));

        return indepNum;
    }

    /**
     * Return the the names of the independent variables from the reader.
     *
     * @param reader    A reader for the CSV data.
     * @param tokenizer The tokenizer used to parse the data with the correct delimiter
     *                  set.
     * @param numDims   The number of independent variables (dimensions) to the data.
     */
    private static FastTable<Text> readIndepNames(LineNumberReader reader, TextTokenizer tokenizer, int numDims) throws IOException {

        FastTable<Text> indepNames = FastTable.newInstance();

        String aLine = reader.readLine();
        if (aLine == null)
            throw new IOException(format(RESOURCES.getString("EOFErrMsg"), reader.getLineNumber()));

        tokenizer.setText(aLine);
        try {
            for (int i = 0; i < numDims; ++i) {
                Text token = tokenizer.nextToken();
                indepNames.add(token.trim());
            }
        } catch (NoSuchElementException e) {
            throw new IOException(format(RESOURCES.getString("csvUnexpNumDimsMsg"), numDims, reader.getLineNumber()), e);
        }

        return indepNames;
    }

    /**
     * Write out a list of Float64Table objects to a CSV formatted file. A CSV file
     * requires that all the tables have the same independent variables with the same
     * number of breakpoints in each!
     *
     * @param output Reference to a file to write the tables to.
     * @param tables A list of tables to be written to the file.
     */
    private static void writeCSVFile(BufferedWriter output, FTableList tables) throws IOException {
        GeneralFormat fmt = new GeneralFormat(Locale.US);

        Float64Table table0 = tables.get(0);
        int numDims = table0.getDimensions();

        // Write out the number of breakpoints in each dimension.
        for (int i = 0; i < numDims; ++i) {
            if (i != 0)
                output.write(DELIMITER);
            output.write(fmt.format(table0.getNumBreakpoints(i)));
        }
        output.newLine();

        // Write out the name of each independent variable.
        for (int i = 0; i < numDims; ++i) {
            Text indepName = table0.getIndepName(i);
            indepName.print(output);
            output.write(DELIMITER);
        }

        // Write out the name of the dependent variables.
        for (Float64Table table : tables) {
            Text.valueOf(table.getName()).print(output);
            output.write(DELIMITER);
        }
        output.newLine();

        // Get the number of elements in each of the tables.
        int numElements = table0.size();

        // Write out all the independent and dependent values (one line at a time).
        FastTable<Integer64> pos = FastTable.newInstance();
        for (int i = 0; i < numDims; ++i)
            pos.add(Integer64.ZERO);

        for (int i = 0; i < numElements; ++i) {
            for (int j = 0; j < numDims; ++j) {
                // Write out the independent values.
                Float64Vector breakpoints = table0.getBreakpoints(j);
                double value = breakpoints.getValue(pos.get(j).intValue());
                output.write(fmt.format(value));
                output.write(DELIMITER);
            }

            // Write out the dependent values.
            for (Float64Table table : tables) {
                double value = table.getValue(pos);
                output.write(fmt.format(value));
                output.write(DELIMITER);
            }

            // Go to the next line.
            output.newLine();

            // Change position in the tables.
            int k = numDims - 1;
            pos.set(k, pos.get(k).plus(Integer64.ONE));         //  ++pos[k]
            while (k > 0 && pos.get(k).intValue() >= table0.getNumBreakpoints(k)) {
                pos.set(k, Integer64.ZERO);
                --k;
                pos.set(k, pos.get(k).plus(Integer64.ONE));     //  ++pos[k]
            }
        }
    }

    /**
     * Check the tables that are to be written out to make sure that they are consistent.
     * CSV tables must all have exactly the same breakpoints.
     *
     * @param tables A list of tables that we are writing out.
     * @throws IOException if the tables are not consistent.
     */
    private static void checkTables(FTableList tables) throws IOException {

        // Check for consistant tables.
        int numDims = tables.get(0).getDimensions();

        for (Float64Table table : tables) {
            if (table.getDimensions() != numDims)
                throw new ArrayIndexOutOfBoundsException(RESOURCES.getString("csvTableDimErrMsg"));

            for (int j = 0; j < numDims; ++j) {

                CharSequence indName0 = tables.get(0).getIndepName(j);
                if (!table.getIndepName(j).equals(indName0))
                    throw new ArrayIndexOutOfBoundsException(RESOURCES.getString("csvBreakpointErrMsg"));

                Float64Vector breakpoints0 = tables.get(0).getBreakpoints(j);
                if (!table.getBreakpoints(j).equals(breakpoints0)) {
                    throw new ArrayIndexOutOfBoundsException(RESOURCES.getString("csvBreakpointErrMsg"));
                }
            }
        }
    }

    /**
     * Write out any database notes as file level comments to the CSV file.
     *
     * @param output A writer used to output the data.
     * @param tables The table database we are writing out.
     */
    private static void writeFileComments(BufferedWriter output, FTableList tables)
            throws IOException {
        int numNotes = tables.numberOfNotes();
        if (numNotes == 0) {

            // Write out some generic table comments into the header lines.
            output.write("\"");
            tables.toText().print(output);
            output.write("\"");
            output.newLine();
            output.write(",");
            output.newLine();
            output.write("\"Date:, ");
            Date theDate = new Date();
            output.write(DateFormat.getDateInstance(DateFormat.SHORT).format(theDate));
            output.write(", ");
            output.write(DateFormat.getTimeInstance().format(theDate));
            output.write("\"");
            output.newLine();
            output.write(",");
            output.newLine();

        } else {
            // Write out the table notes as the header of the file.
            if (numNotes > 4)
                numNotes = 4;

            for (int i = 0; i < numNotes; ++i) {
                Text note = tables.getNote(i);
                if (!note.startsWith("\""))
                    output.write("\"");

                output.write(note.toString());
                if (!note.endsWith("\""))
                    output.write("\"");

                output.newLine();
            }
            for (int i = 4 - numNotes; i > 0; --i) {
                output.write(",");
                output.newLine();
            }
        }
    }

}
