/*
 * FTableList    -- A list of Float64Table objects with some additional
 *                      information related to the collection of tables.
 *
 * Copyright (C) 1999-2024, by Joseph A. Huwaldt.
 * All rights reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *   
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 * Or visit:  http://www.gnu.org/licenses/lgpl.html
 */
package jahuwaldt.js.tables;

import java.text.MessageFormat;
import java.util.Arrays;
import java.util.List;
import java.util.Objects;
import static java.util.Objects.isNull;
import static java.util.Objects.requireNonNull;
import java.util.ResourceBundle;
import javolution.context.ObjectFactory;
import javolution.text.Text;
import javolution.util.FastTable;
import javolution.xml.XMLFormat;
import javolution.xml.stream.XMLStreamException;
import org.jscience.mathematics.vector.*;

/**
 * A random access list of {@link Float64Table} objects with real-time behavior. This list
 * contains both the tables and an optional set of "database level" notes.
 * 
 * <p> Modified by: Joseph A. Huwaldt </p>
 * 
 * @author Joseph A. Huwaldt, Date: January 6, 2000
 * @version January 15, 2024
 */
public class FTableList extends FastTable<Float64Table> {

    private static final long serialVersionUID = 3272329698640710146L;

    /**
     * The resource bundle for this package.
     */
    static final ResourceBundle RESOURCES = Float64Table.RESOURCES;
    
    /**
     * A list of notes stored with the list of tables.
     */
    private FastTable<Text> _notes;

    //-----------------------------------------------------------------------------
    /**
     * Returns a <code>FTableList</code> object that contains the specified table.
     *
     * @param table A single table to create this list from.
     * @return The list containing the specified table.
     */
    public static FTableList valueOf(Float64Table table) {
        requireNonNull(table, MessageFormat.format(RESOURCES.getString("paramNullErr"), "table"));
        FTableList list = newInstance();
        list.add(table);
        return list;
    }

    /**
     * Returns a <code>FTableList</code> object from an list of <code>Float64Table</code>
     * objects.
     *
     * @param tables A list of tables to create this list from.
     * @return The list containing the specified tables.
     */
    public static FTableList valueOf(List<Float64Table> tables) {
        requireNonNull(tables, MessageFormat.format(RESOURCES.getString("paramNullErr"), "tables"));
        FTableList list = newInstance();
        list.addAll(tables);
        return list;
    }

    /**
     * Returns a <code>FTableList</code> object from an array of <code>Float64Table</code>
     * objects.
     *
     * @param tables An array of tables to create this list from.
     * @return The list containing the specified tables.
     */
    public static FTableList valueOf(Float64Table... tables) {
        requireNonNull(tables, MessageFormat.format(RESOURCES.getString("paramNullErr"), "tables"));
        FTableList list = newInstance();
        list.addAll(Arrays.asList(tables));
        return list;
    }

    //-----------------------------------------------------------------------------
    /**
     * Tests to see if a table with the specified name exists in the list.
     *
     * @param name The name we are looking for in the list.
     * @return true if the named table is contained in this list.
     */
    public boolean containsName(CharSequence name) {
        return getIndexFromName(name) >= 0;
    }

    /**
     * Returns the table with the specified name from the list.
     *
     * @param name The name of the table we are looking for in the list.
     * @return The table matching the specified name. If the specified table name isn't
     *         found in the list, then null is returned.
     */
    public Float64Table get(CharSequence name) {

        Float64Table table = null;
        int index = getIndexFromName(name);
        if (index >= 0)
            table = this.get(index);

        return table;
    }

    /**
     * Removes the table with the specified name from this list. Shifts any subsequent
     * values to the left (subtracts one from their indices). Returns the table that was
     * removed from the list.
     *
     * @param name The name of the table that is to be removed.
     * @return The table matching the specified name. If the specified table isn't found
     *         (nothing is removed), then null is returned.
     */
    public Float64Table remove(CharSequence name) {

        Float64Table table = null;
        int index = getIndexFromName(name);
        if (index >= 0)
            table = this.remove(index);

        return table;
    }

    /**
     * Returns the number of notes associated with this list of tables.
     *
     * @return The number of notes associated with this list of tables.
     */
    public int numberOfNotes() {
        if (_notes == null)
            return 0;
        return _notes.size();
    }

    /**
     * Add a new note to the list of notes associated with this list of tables.
     *
     * @param note The new note to add to this list of tables.
     */
    public void addNote(CharSequence note) {
        requireNonNull(note);
        if (isNull(_notes))
            _notes = FastTable.newInstance();
        _notes.add(Text.valueOf(note));
    }

    /**
     * Add all the notes in a specified list of notes to the list of notes associated with
     * this list of tables.
     *
     * @param noteList The list of new notes to add to this list of tables.
     */
    public void addAllNotes(List<? extends CharSequence> noteList) {
        requireNonNull(noteList, MessageFormat.format(RESOURCES.getString("paramNullErr"), "noteList"));
        for (CharSequence note : noteList)
            addNote(note);
    }

    /**
     * Return a specified note from the list of notes associated with this list of tables.
     *
     * @param index The index of the note to retrieve.
     * @return The note at the specified index in the list of notes.
     */
    public Text getNote(int index) {
        if (isNull(_notes))
            throw new ArrayIndexOutOfBoundsException(index);
        return _notes.get(index);
    }

    /**
     * Return all the notes associated with this list of tables as an unmodifiable list.
     * If there are no notes associated with this list, then <code>null</code> is
     * returned.
     *
     * @return All the notes associated with this list of tables
     */
    public List<Text> getAllNotes() {
        if (isNull(_notes))
            _notes = FastTable.newInstance();
        return _notes.unmodifiable();
    }

    /**
     * Remove a specified note from the list of notes associated with this list of tables.
     *
     * @param index The index of the note to remove.
     * @return The removed note.
     */
    public Text removeNote(int index) {
        if (isNull(_notes))
            throw new ArrayIndexOutOfBoundsException(index);
        return _notes.remove(index);
    }

    /**
     * Removes all notes associated with this list of tables.
     */
    public void clearNotes() {
        if (!isNull(_notes))
            _notes.clear();
    }

    /**
     * Compares the specified object with this list of <code>Float64Table</code> objects
     * for equality. Returns true if and only if both collections contain equal values in
     * the same order.
     *
     * @param obj the object to compare with.
     * @return <code>true</code> if this list is identical to that list;
     *         <code>false</code> otherwise.
     */
    @Override
    public boolean equals(Object obj) {
        if (this == obj)
            return true;
        if ((obj == null) || (obj.getClass() != this.getClass()))
            return false;

        FTableList that = (FTableList)obj;
        if (!Objects.equals(this._notes, that._notes))
                return false;
        if (!this._notes.equals(that._notes))
            return false;

        return super.equals(obj);
    }

    /**
     * Returns the hash code for this <code>FTableList</code>.
     *
     * @return the hash code value.
     */
    @Override
    public int hashCode() {
        int hash = 7;

        hash = hash * 31 + (_notes == null ? 0 : _notes.hashCode());
        hash = hash * 31 + super.hashCode();

        return hash;
    }

    //-----------------------------------------------------------------------------
    /**
     * Returns the index into the list of the table with the specified name. If the name
     * is not found, -1 is returned. Make sure and check for this!
     */
    private int getIndexFromName(CharSequence name) {
        Text nameText = Text.valueOf(name);

        int result = -1;
        int size = this.size();
        for (int i = 0; i < size; ++i) {
            Float64Table table = this.get(i);
            Text tblName = Text.valueOf(table.getName());
            if (tblName.equals(nameText)) {
                result = i;
                break;
            }
        }

        return result;
    }

    /**
     * Holds the default XML representation. For example:
     * <pre>
     *    <FTableList>
     *        <Tables>
     *            <Float64Table name="table1">
     *                ...
     *            </Float64Table>
     *            <Float64Table name="table2">
     *                ...
     *            </Float64Table>
     *        </Tables>
     *        <Notes>
     *            <Text value="This is a sample note." />
     *        </Notes>
     * </pre>
     */
    protected static final XMLFormat<FTableList> XML = new XMLFormat<FTableList>(FTableList.class) {

        @Override
        public FTableList newInstance(Class<FTableList> cls, InputElement xml) throws XMLStreamException {
            return FTableList.newInstance();
        }

        @Override
        public void read(InputElement xml, FTableList T) throws XMLStreamException {

            //  Read the data from the XML element.
            FastTable<Float64Table> tables = xml.get("Tables", FastTable.class);
            FastTable<Text> notes = xml.get("Notes", FastTable.class);

            if (xml.hasNext()) {
                FastTable.recycle(tables);
                FastTable.recycle(notes);
                throw new XMLStreamException("Too many elements");
            }

            //  Fill in the list of tables.
            T.addAll(tables);
            T.addAllNotes(notes);

            FastTable.recycle(tables);
            FastTable.recycle(notes);
        }

        @Override
        public void write(FTableList T, OutputElement xml) throws XMLStreamException {

            //  Write out the tables 1st.
            FastTable<Float64Table> list = FastTable.newInstance();
            list.addAll(T);
            xml.add(list, "Tables", FastTable.class);
            FastTable.recycle(list);

            //  Now write out the notes.
            if (T._notes != null)
                xml.add(T._notes, "Notes", FastTable.class);

        }
    };

    //////////////////////
    // Factory Creation //
    //////////////////////
    /**
     * Creates an empty FTableList object that can be used to store a list of
     * {@link Float64Table} objects and any associated notes.
     */
    protected FTableList() { }

    /**
     * Returns a new, preallocated or recycled <code>FTableList</code> instance (on the
     * stack when executing in a <code>StackContext</code>) that can store a list of
     * {@link Float64Table} objects and any associated notes.
     *
     * @return
     */
    @SuppressWarnings("unchecked")
    public static FTableList newInstance() {
        FTableList o = FACTORY.object();
        if (o._notes == null)
            o._notes = FastTable.newInstance();
        return o;
    }

    private static final ObjectFactory<FTableList> FACTORY = new ObjectFactory<FTableList>() {
        @Override
        protected FTableList create() {
            return new FTableList();
        }

        @Override
        protected void cleanup(FTableList obj) {
            obj.reset();
            obj._notes.reset();
        }
    };

    /**
     * Tests the methods in this class.
     *
     * @param args Command-line arguments (ignored).
     */
    public static void main(String args[]) {
        System.out.println("Testing FTableList:");

        //  Test 2 tables.
        Float64Vector indep1 = Float64Vector.valueOf(0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10);
        Float64Vector vector1 = Float64Vector.valueOf(0, 10, 20, 30, 40, 50, 60, 70, 80, 90, 100);
        Float64Table table1 = Float64Table.valueOf(Text.valueOf("table1"), Text.valueOf("indep1"), indep1, vector1);

        Float64Vector indep2 = Float64Vector.valueOf(100, 200, 300);
        List<Float64Vector> dep2D = FastTable.newInstance();
        dep2D.add(vector1);
        dep2D.add(Float64Vector.valueOf(0, 100, 200, 300, 400, 500, 600, 700, 800, 900, 1000));
        dep2D.add(Float64Vector.valueOf(0, 1000, 2000, 3000, 4000, 5000, 6000, 7000, 8000, 9000, 10000));
        List<Text> indepNames = new FastTable<>();
        indepNames.add(Text.valueOf("indep2"));
        indepNames.add(Text.valueOf("indep1"));
        Float64Table table2 = Float64Table.valueOf(Text.valueOf("table2"), indepNames, indep2, indep1, dep2D);

        //  Create a table list.
        FTableList list = FTableList.newInstance();
        list.add(table1);
        list.add(table2);
        System.out.println("table1 = " + table1);
        System.out.println("table2 = " + table2);
        System.out.println("list = " + list);

        System.out.println("\nAdding notes to the list:");
        String note1 = "This is note #1";
        Text note2 = Text.valueOf("This is note #2");
        list.addNote(note1);
        list.addNote(note2);
        System.out.println("list.getAllNotes() = " + list.getAllNotes());

        //  Write out XML data.
        try {
            System.out.println();
            javolution.xml.XMLObjectWriter writer = javolution.xml.XMLObjectWriter.newInstance(System.out);
            writer.setIndentation("    ");
            writer.write(list, "FTableList", FTableList.class);
            writer.flush();

        } catch (Exception e) {
            e.printStackTrace();
        }

    }
}
