/*
*   POST2TableReader -- A class that can read and write POST II table (*.dat) multi-table files.
*
*   Copyright (C) 2009-2024, by Joseph A. Huwaldt.
*   All rights reserved.
*
*   This library is free software; you can redistribute it and/or
*   modify it under the terms of the GNU Lesser General Public
*   License as published by the Free Software Foundation; either
*   version 2 of the License, or (at your option) any later version.
*
*   This library is distributed in the hope that it will be useful,
*   but WITHOUT ANY WARRANTY; without even the implied warranty of
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
*   Library General Public License for more details.
*
*  You should have received a copy of the GNU Lesser General Public License
*  along with this program; if not, write to the Free Software
*  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*  Or visit:  http://www.gnu.org/licenses/lgpl.html
*/
package jahuwaldt.js.tables;

import static java.io.StreamTokenizer.TT_EOF;
import static java.io.StreamTokenizer.TT_NUMBER;
import static java.io.StreamTokenizer.TT_WORD;
import java.io.*;
import java.text.DateFormat;
import java.text.MessageFormat;
import static java.text.MessageFormat.format;
import java.util.Date;
import java.util.List;
import java.util.Locale;
import static java.util.Objects.requireNonNull;
import javolution.text.Text;
import javolution.util.FastMap;
import javolution.util.FastTable;
import org.jscience.mathematics.number.Float64;
import org.jscience.mathematics.number.Integer64;
import org.jscience.mathematics.vector.Float64Vector;

/**
 * A class that provides methods for reading or writing a POST II table (*.dat) formatted
 * multi-table file. These files are used by the POST II trajectory simulation system to
 * store aerodynamic and propulsion data among other things. See POST II documentation for
 * more information.
 *
 * <p> Modified by: Joseph A. Huwaldt </p>
 *
 * @author Joseph A. Huwaldt, Date: September 7, 2008
 * @version January 15, 2024
 */
public class POST2TableReader implements FTableReader {

    //  Debug flag.
    private static final boolean DEBUG = false;

    //  The preferred file extension for files of this reader's type.
    private static final String EXTENSION = "dat";

    private static final int MONOVAR = 1;
    private static final int MULTI = 2;
    private static final int MULTISUB = 3;

    //  A map of sub-table numbers for each table
    private static final FastMap<CharSequence,Integer> SUB_TBL_NUM_MAP = FastMap.newInstance();


    /**
     * Returns a string representation of the object. This will return a brief description
     * of the format read by this reader.
     */
    @Override
    public String toString() {
        return RESOURCES.getString("p2Description");
    }

    /**
     * Returns the preferred file extension (not including the ".") for files of this
     * reader's type.
     *
     * @return The file extension "dat".
     */
    @Override
    public String getExtension() {
        return EXTENSION;
    }

    /**
     * Method that determines if this reader can read data from the specified input
     * stream. Will return FTableReader.YES if the stream is recognized as an POST table
     * file.
     *
     * @param name  The name of the file.
     * @param input An input stream containing the data to be read. Any methods that read
     *              from this stream must first set a mark and then reset back to that
     *              mark before the method returns (even if it returns with an exception).
     * @return FTableReader.NO if the file is not recognized at all or FTableReader.YES if
     *         the file appears to contain POST tables.
     * @throws java.io.IOException if there was a problem reading form the file.
     */
    @Override
    public int canReadData(String name, BufferedInputStream input) throws IOException {
        requireNonNull(input, MessageFormat.format(RESOURCES.getString("paramNullErr"), "input"));
        
        //  Mark the current position in the input stream so that we can return to it.
        input.mark(10240);  //  10KB buffer

        try {
            //  Create a buffer to hold the data read in from the file.
            byte[] buffer = new byte[10238];

            //  Read in up to 10k worth of data.
            int byteCount = input.read(buffer, 0, 10238);
            if (byteCount < 0)  return NO;

            //  Turn the buffer into a (potentially 10k long) string.
            String str = new String(buffer, 0, byteCount);

            //  Convert the string into a line number reader.
            LineNumberReader reader = new LineNumberReader(new StringReader(str));

            //  Look for "monovar" or "multi" or "multisub" on a line that contains an equal sign.
            Text aLine = findHeaderLine(reader);
            if (aLine == null)
                return NO;

        } catch (Exception e) {
            return NO;

        } finally {
            input.reset();
        }

        return YES;
    }

    /**
     * Returns true. This class can write data to an POST II table file.
     *
     * @return Always returns true.
     */
    @Override
    public boolean canWriteData() {
        return true;
    }

    /**
     * Compares this object with the specified object for order based on the
     * <code>toString().compareTo(o.toString())</code> method. Returns a negative integer,
     * zero, or a positive integer as this object is less than, equal to, or greater than
     * the specified object.
     */
    @Override
    public int compareTo(FTableReader o) {
        requireNonNull(o, MessageFormat.format(RESOURCES.getString("paramNullErr"), "o"));
        return this.toString().compareTo(o.toString());
    }

    /**
     * Method that reads in POST II table formatted data from the specified input stream
     * and returns that data as an {@link FTableList} object.
     *
     * @param input An input stream containing the POST II table formatted data.
     * @return An {@link FTableList} object that contains the data read in from the
     *         specified stream.
     * @throws IOException If there is a problem reading the specified stream.
     */
    @Override
    public FTableList read(InputStream input) throws IOException {
        requireNonNull(input, MessageFormat.format(RESOURCES.getString("paramNullErr"), "input"));

        //  POST II files are required to be in ASCII with U.S. style number formatting.
        //  Get the default locale and save it.  Then change the default locale to US.
        Locale defLocale = Locale.getDefault();
        Locale.setDefault(Locale.US);

        try {
            // Create a reader to translate the input stream.
            BufferedReader r = new BufferedReader(new InputStreamReader(input));

            //  Create a tokenizer for the stream.
            StreamTokenizer tokenizer = new StreamTokenizer(r);
            tokenizer.lowerCaseMode(true);
            tokenizer.commentChar('*');
            tokenizer.commentChar('/');
            tokenizer.whitespaceChars(',', ',');
            tokenizer.whitespaceChars('+', '+');
            tokenizer.wordChars('_', '_');
            tokenizer.wordChars('-', '-');

            FTableList db = readStream(tokenizer);

            return db;

        } finally {
            //  Restore the default locale.
            Locale.setDefault(defLocale);
        }
    }

    /**
     * Method for writing out all the data stored in the specified list of
     * {@link Float64Table} objects to the specified output stream in POST II table
     * format. This method is not implemented yet and will throw an exception if called.
     *
     * @param output The output stream to which the data is to be written.
     * @param tables The list of <code>Float64Table</code> objects to be written out.
     * @throws IOException If there is a problem writing to the specified stream.
     */
    @Override
    public void write(OutputStream output, FTableList tables) throws IOException {
        requireNonNull(output, MessageFormat.format(RESOURCES.getString("paramNullErr"), "output"));
        requireNonNull(tables, MessageFormat.format(RESOURCES.getString("paramNullErr"), "tables"));
        if (tables.isEmpty())
            return;

        //  POST II files are required to be in ASCII with U.S. style number formatting.
        //  Get the default locale and save it.  Then change the default locale to US.
        Locale defLocale = Locale.getDefault();
        Locale.setDefault(Locale.US);

        try {
            // Get a reference to the output stream.
            PrintWriter writer = new PrintWriter(output);

            //  Write out the comments.
            writeFileComments(writer, tables);

            //  Write out the list of tables.
            for (Float64Table table : tables) {
                List<Text> notes = table.getAllNotes();
                for (Text note : notes) {
                    writer.print("c ");
                    note.println(writer);
                }
                writeTable(writer, table);
            }

            //  Flush the output.
            writer.flush();

        } finally {
            //  Restore the default locale.
            Locale.setDefault(defLocale);

            //  Clear the list of sub-table numbers.
            SUB_TBL_NUM_MAP.clear();
        }

    }

    /**
     * Read in a POST II table formatted multi-table stream and create a new table
     * database with the data.
     *
     * @param tokenizer The tokenizer for the stream.
     * @return A table database containing the tables read in from the file.
     */
    private static FTableList readStream( StreamTokenizer tokenizer ) throws IOException {

        // Create an empty table database.
        FTableList db = FTableList.newInstance();

        //  Create a list of multivariant table data sets.
        //  Outside key = table name.
        //  Inside key = sub-table index number.
        FastMap<String,FastMap<Integer,Integer>> multiTblTypes = FastMap.newInstance();
        FastMap<String,FastMap<Integer,String>> multiTblIndeps = FastMap.newInstance();
        FastMap<String,FastMap<Integer,FastTable<Double>>> multiTblsX = FastMap.newInstance();
        FastMap<String,FastMap<Integer,FastTable<Double>>> multiTblsY = FastMap.newInstance();

        //  Start reading in tokens.
        int ttype = tokenizer.nextToken();
        while (ttype != TT_EOF) {

            //  Get the table name.
            if (ttype != TT_WORD)
                throw new IOException(format(RESOURCES.getString("p2MissingTbleNameErr"), tokenizer.lineno()));
            String tblName = tokenizer.sval;
            if (tblName == null || tblName.length() == 0)
                throw new IOException(format(RESOURCES.getString("p2CantParseTblNameErr"), tokenizer.lineno()));

            //  Skip the equals sign.
            ttype = tokenizer.nextToken();
            if (ttype != '=')   throw new IOException(format(RESOURCES.getString("p2MissingEqualSignErr"), tokenizer.lineno()));

            //  Skip the table label.
            ttype = tokenizer.nextToken();
            if (ttype != TT_WORD && ttype != '\'' && ttype != '"')
                throw new IOException(format(RESOURCES.getString("p2CantParseLabelErr"), tokenizer.lineno()));

            //  Get the table type.
            ttype = tokenizer.nextToken();
            if (ttype != TT_WORD)
                throw new IOException(format(RESOURCES.getString("p2MissingTbleTypeErr"), tokenizer.lineno()));
            int tblType=0;
            String token = tokenizer.sval.trim().toLowerCase();
            switch(token) {
                case "monovar":
                    tblType = MONOVAR;
                    break;
                case "multisub":
                    tblType = MULTISUB;
                    break;
                case "multi":
                    tblType = MULTI;
                    break;
                default:
                    throw new IOException(format(RESOURCES.getString("p2UnknownTypeErr"), token, tokenizer.lineno()));
            }

            //  Get the independent variable name for this table.
            ttype = tokenizer.nextToken();
            if (ttype != TT_WORD)
                throw new IOException(format(RESOURCES.getString("p2MissingIndepVarErr"), tokenizer.lineno()));
            String indepName = tokenizer.sval;

            //  Get the sub-table number.
            ttype = tokenizer.nextToken();
            if (ttype != TT_NUMBER) throw new IOException(format(RESOURCES.getString("p2MissingSubTblNumErr"), tokenizer.lineno()));
            int subTblNum = (int)tokenizer.nval;

            //  Skip the interpolation type.
            TRUtils.nextWord(tokenizer);

            //  Skip the extrapolation type.
            TRUtils.nextWord(tokenizer);

            if (DEBUG) {
                System.out.println("tblName = " + tblName + ", tblType = " + tblType + ", indepName = " + indepName +
                                    ", subTblNum = " + subTblNum);
            }

            //  Store the data so far.

            //  Store the table type.
            FastMap<Integer,Integer> typeMap = multiTblTypes.get(tblName);
            if (typeMap == null) {
                typeMap = FastMap.newInstance();
                multiTblTypes.put(tblName, typeMap);
            }
            typeMap.put(subTblNum, tblType);

            //  Store the table independent variable.
            FastMap<Integer,String> indepMap = multiTblIndeps.get(tblName);
            if (indepMap == null) {
                indepMap = FastMap.newInstance();
                multiTblIndeps.put(tblName, indepMap);
            }
            indepMap.put(subTblNum, indepName);


            //  Start reading in data pairs.
            FastMap<Integer,FastTable<Double>> tblXs = multiTblsX.get(tblName);
            if (tblXs == null) {
                tblXs = FastMap.newInstance();
                multiTblsX.put(tblName, tblXs);
            }
            FastTable<Double> xValues = FastTable.newInstance();
            tblXs.put(subTblNum,xValues);
            FastMap<Integer,FastTable<Double>> tblYs = multiTblsY.get(tblName);
            if (tblYs == null) {
                tblYs = FastMap.newInstance();
                multiTblsY.put(tblName, tblYs);
            }
            FastTable<Double> yValues = FastTable.newInstance();
            tblYs.put(subTblNum,yValues);

            //  Read until an exception occurs (until we get something that is not a number).
            while (true) {
                double xx;

                //  Read in the X value.
                try {
                    xx = TRUtils.nextNumber(tokenizer);
                } catch (IOException e) {
                    //  If we get a bad X, then continue on.
                    ttype = tokenizer.ttype;
                    break;
                }

                //  If we get a good X value, then we require a good Y value to go with it.
                double yy = TRUtils.nextNumber(tokenizer);

                xValues.add(xx);
                yValues.add(yy);
            }
            if (DEBUG) {
                System.out.println("xValues = " + xValues);
                System.out.println("yValues = " + yValues);
            }

        }   //  end while (ttype != TT_EOF)


        //  Loop over all the tables read in and convert them to Float64Table objects.
        int numTables = multiTblTypes.size();
        String[] tblNames = new String[numTables];
        multiTblTypes.keySet().toArray(tblNames);       //  Get the table names.
        for (int ii = 0; ii < numTables; ++ii) {
            String tblName = tblNames[ii];
            if (DEBUG)
                System.out.println("Processing table " + tblName);

            //  Get the maps associated with this table.
            FastMap<Integer,Integer> typeMap = multiTblTypes.get(tblName);
            FastMap<Integer,String> indepMap = multiTblIndeps.get(tblName);
            FastMap<Integer,FastTable<Double>> tblXs = multiTblsX.get(tblName);
            FastMap<Integer,FastTable<Double>> tblYs = multiTblsY.get(tblName);

            //  Extract the dimensions and breakpoints.
            FastTable<Float64Vector> indepArrs = FastTable.newInstance();
            FastTable<String> indepNames = FastTable.newInstance();
            int numDims = 1;
            int subTblNum = 0;
            indepNames.add(indepMap.get(subTblNum));
            indepArrs.add(convertDoubles2Float64s(tblXs.get(subTblNum)));
            while(typeMap.get(subTblNum) == MULTI) {
                ++numDims;
                subTblNum = tblYs.get(subTblNum).get(0).intValue();
                indepNames.add(indepMap.get(subTblNum));
                indepArrs.add(convertDoubles2Float64s(tblXs.get(subTblNum)));
            }

            //  Create an initially empty table.
            Float64Table table = Float64Table.valueOf(tblName, indepNames, indepArrs, null);

            //  Recursively convert the data into the Float64Table format.
            FastTable<Integer64> pos = FastTable.newInstance();
            for (int i=0; i < numDims; ++i) pos.add(Integer64.ZERO);
            rConvertData(numDims, 0, 0, pos, tblYs, table);

            //  Add the new table to the database of tables.
            db.add(table);

            //  Clean up before looping.
            FastTable.recycle(indepArrs);
            FastTable.recycle(indepNames);
            FastTable.recycle(pos);
        }

        //  Clean up before leaving.
        FastMap.recycle(multiTblTypes);
        FastMap.recycle(multiTblIndeps);
        FastMap.recycle(multiTblsX);
        FastMap.recycle(multiTblsY);

        return db;
    }

    /**
     * Recursively store the data from the specified maps into the specified table.
     *
     * @param dim       The current dimension being converted.
     * @param subTblNum The current sub-table number.
     * @param indep     The independent position vector being built up.
     * @param tblYs     The table Y values: key = subTblNum, value = list of Y values.
     * @param table     The table being filled out.
     */
    private static void rConvertData(int numDims, int dim, int subTblNum, FastTable<Integer64> indep,
                                        FastMap<Integer,FastTable<Double>> tblYs, Float64Table table ) throws IOException {

        if (dim == numDims-1) {
            //  Reached the lowest dimension.
            //  Copy in the array.
            int count = 0;
            FastTable<Double> yValues = tblYs.get(subTblNum);
            if (DEBUG)
                System.out.println("    subTblNum = " + subTblNum + ", numPoints = " + yValues.size() + ", numBreakpoints = " + table.getNumBreakpoints(dim));
            if (yValues.size() != table.getNumBreakpoints(dim))
                throw new IOException(format(RESOURCES.getString("p2VarBreakpointsErr"), table.getName()));

            for (Double value : yValues) {
                indep.set(numDims-1, Integer64.valueOf(count));     //  indep[numDims-1] = count;
                table.set(indep, value);
                ++count;
            }

        } else {
            // Loop over all the breakpoints in this dimension.
            FastTable<Double> yValues = tblYs.get(subTblNum);
            for ( int i = 0; i < yValues.size(); ++i ) {

                // Fill in the required value for indep[] at this dimension.
                indep.set(dim, Integer64.valueOf(i));

                // Go on to the next lower dimension.
                rConvertData( numDims, dim + 1, yValues.get(i).intValue(), indep, tblYs, table );
            }
        }
    }


    /**
     * Convert a list of Double values to a Float64Vector.
     */
    private static Float64Vector convertDoubles2Float64s(List<Double> list) {
        FastTable<Float64> values = FastTable.newInstance();
        for (Double value : list)
            values.add(Float64.valueOf(value));

        Float64Vector vector = Float64Vector.valueOf(values);
        FastTable.recycle(values);
        return vector;
    }

    /**
     * Method that reads input from the the specified stream expecting the 1st non-comment
     * line to be a line with table related key words on it ("monovar", or "multisub" or
     * "multi" following an equal sign). If a table header is found, it is returned,
     * otherwise <code>null</code> is returned.
     */
    private Text findHeaderLine(LineNumberReader reader) throws IOException {

        Text aLine = readLineNoComments(reader);
        while (aLine != null) {
            Text aLineLC = aLine.toLowerCase();
            int pos = aLine.indexOf("=");
            if (pos > 0) {
                //  Does the line contain any of our table related key words?
                ++pos;
                if (aLineLC.indexOf("monovar", pos) > 0 || aLineLC.indexOf("multi", pos) > 0)
                    return aLine.trim();
            }

            //  Read the next line.
            aLine = readLineNoComments(reader);
        }

        return null;
    }

    /**
     * Reads a line of text from the supplied file reader stripping out any lines that are
     * commented out. Returns <code>null</code> if the end-of-file is reached without
     * finding any "non-comment" text.
     */
    private static Text readLineNoComments(LineNumberReader reader) throws IOException {

        String aLine = reader.readLine();
        while (aLine != null &&
                (aLine.startsWith("c") || aLine.startsWith("/") || aLine.startsWith("*") || aLine.trim().length() == 0) )
            aLine = reader.readLine();
        if (aLine == null)  return null;

        //  Strip off anything after the "/" comment character.
        Text text = Text.valueOf(aLine);
        int indx = text.indexOf("/");
        if (indx >= 0)
            text = text.subtext(0,indx);

        return text;
    }


    /**
     * Write out the file level comments to the specified writer.
     */
    private static void writeFileComments(PrintWriter out, FTableList tables) throws IOException {
        List<Text> notes = tables.getAllNotes();
        for ( Text note : notes ) {
            out.print( "c " );
            note.println(out);
        }

        //  Write out some generic table comments.
        out.print("c Date and time written: ");
        out.println(  DateFormat.getDateTimeInstance().format( new Date() ) );
        out.println("c");
    }

    /**
     * Write out a single table to a POST II formatted output stream.
     *
     * @param out   Reference to a PrintWriter to write the specified table to.
     * @param table The table to be written out.
     */
    private static void writeTable(PrintWriter out, Float64Table table) throws IOException {

        //  Recursively write out the table.
        FastTable<Integer64> pos = FastTable.newInstance();
        int numDims = table.getDimensions();
        for (int i=0; i < numDims; ++i) pos.add(Integer64.ZERO);
        rWriteTable(out, table, 0, 0, pos);

        //  Clean up before leaving.
        FastTable.recycle(pos);
    }

    /**
     * Recursively write out the table.
     */
    private static void rWriteTable(PrintWriter out, Float64Table table, int dim, int subTblNum, List<Integer64> indep) {
        int numDims = table.getDimensions();
        if (dim == numDims - 1) {
            //  Reached the lowest dimension.
            //  Write out the table at this level.

            //  Write out the header line.
            String tblName = table.getName().toString();
            out.print(" ");
            out.print(tblName);
            out.print(" = ");
            out.print(tblName);
            out.print(", ");
            if (numDims > 1)
                out.print("multisub");
            else
                out.print("monovar");
            out.print(", ");
            out.print(table.getIndepName(dim).toString());
            out.print(", ");
            out.print(subTblNum);
            out.print(", ");
            out.println("lin_inp, noxt,");

            //  Write out the pairs of X, Y points.
            Float64Vector brkPoints = table.getBreakpoints(dim);
            int numIndep = brkPoints.getDimension();
            for (int i=0; i < numIndep; ++i) {
                out.print("        ");
                out.print(brkPoints.getValue(i));
                out.print(", ");
                indep.set(numDims-1, Integer64.valueOf(i));     //  indep[numDims-1] = i;
                out.print(table.getValue(indep));
                out.println(",");
            }

        } else {
            //  Write out the header line.
            String tblName = table.getName().toString();
            out.print(" ");
            out.print(tblName);
            out.print(" = ");
            out.print(tblName);
            out.print(", multi, ");
            out.print(table.getIndepName(dim).toString());
            out.print(", ");
            out.print(subTblNum);
            out.print(", ");
            out.println("lin_inp, noxt,");

            // Loop over all the breakpoints in this dimension and write out a sub-table number for each.
            Float64Vector brkPoints = table.getBreakpoints(dim);
            FastTable<Integer> subTblNums = FastTable.newInstance();
            int numIndep = brkPoints.getDimension();
            for ( int i = 0; i < numIndep; ++i ) {
                int childSubTblNum = nextSubTableNum(tblName);
                subTblNums.add(childSubTblNum);
                out.print("        ");
                out.print(brkPoints.getValue(i));
                out.print(", ");
                out.print(childSubTblNum);
                out.println(",");
            }

            // Loop over all the breakpoints in this dimension.
            for (int i=0; i < numIndep; ++i) {
                // Fill in the required value for indep[] at this dimension.
                indep.set(dim, Integer64.valueOf(i));

                //  Go on to the next lower dimension.
                rWriteTable(out, table, dim+1, subTblNums.get(i), indep);
            }

            //  Clean up before leaving.
            FastTable.recycle(subTblNums);
        }
    }

    /**
     * Returns the next unused sub-table number for the specified table name.
     */
    private static int nextSubTableNum(CharSequence name) {
        int value = 0;
        Integer last = SUB_TBL_NUM_MAP.get(name);
        if (last != null)
            value = last;
        ++value;
        SUB_TBL_NUM_MAP.put(name, value);
        return value;
    }

}


