/*
 *   SHABPTableReader -- A class that can read S/HABP aerodynamics print files.
 *
 *   Copyright (C) 2009-2024, by Joseph A. Huwaldt.
 *   All rights reserved.
 *   
 *   This library is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU Lesser General Public
 *   License as published by the Free Software Foundation; either
 *   version 2 of the License, or (at your option) any later version.
 *   
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   Library General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *  Or visit:  http://www.gnu.org/licenses/lgpl.html
 */
package jahuwaldt.js.tables;

import static jahuwaldt.js.tables.FTableReader.RESOURCES;
import jahuwaldt.js.util.TextTokenizer;
import java.io.*;
import java.text.MessageFormat;
import static java.text.MessageFormat.format;
import java.util.List;
import java.util.Locale;
import java.util.NoSuchElementException;
import static java.util.Objects.requireNonNull;
import javolution.text.Text;
import javolution.util.FastTable;
import org.jscience.mathematics.number.Float64;
import org.jscience.mathematics.vector.Float64Vector;

/**
 * A class that provides methods for reading an S/HABP aerodynamic print file parsing the
 * data into tables of coefficients.
 * 
 * <p> Modified by: Joseph A. Huwaldt </p>
 * 
 * @author Joseph A. Huwaldt, Date: March 23, 2008
 * @version January 15, 2024
 */
public class SHABPTableReader implements FTableReader {

    private static final boolean DEBUG = false;

    //  The preferred file extension for files of this reader's type.
    private static final String EXTENSION = "out";

    private static final Text DELIMITERS = Text.intern(" ");
    private static final Text SHABPCODE = Text.intern("S/HABP");

    private static final String[] TBL_NAMES = {"ALPHA", "BETA", "CN", "CA", "CMb", "CYb", "CLNb", "CLLb"};

    /**
     * Returns a string representation of the object. This will return a brief description
     * of the format read by this reader.
     */
    @Override
    public String toString() {
        return RESOURCES.getString("shabpDescription");
    }

    /**
     * Returns the preferred file extension (not including the ".") for files of this
     * reader's type.
     *
     * @return The file extension "out".
     */
    @Override
    public String getExtension() {
        return EXTENSION;
    }

    /**
     * Method that determines if this reader can read data from the specified input
     * stream. Will return FTableReader.YES if the stream is recognized as an S/HABP print
     * file.
     *
     * @param name  The name of the file.
     * @param input An input stream containing the data to be read. Any methods that read
     *              from this stream must first set a mark and then reset back to that
     *              mark before the method returns (even if it returns with an exception).
     * @return FTableReader.NO if the file is not recognized at all or FTableReader.YES if
     *         the file appears to be an S/HABP print file.
     */
    @Override
    public int canReadData(String name, BufferedInputStream input) throws IOException {
        requireNonNull(input, MessageFormat.format(RESOURCES.getString("paramNullErr"), "input"));
        
        //  Mark the current position in the input stream so that we can return to it.
        input.mark(10240);  //  10KB buffer

        try {
            //  Create a buffer to hold the data read in from the file.
            byte[] buffer = new byte[10238];

            //  Read in up to 10k worth of data.
            int byteCount = input.read(buffer, 0, 10238);
            if (byteCount < 0)
                return NO;

            //  Turn the buffer into a (potentially 10k long) string.
            String str = new String(buffer, 0, byteCount);

            //  Convert the string into a line number reader.
            LineNumberReader reader = new LineNumberReader(new StringReader(str));

            //  Look for the "S/HABP" sequence.
            String aLine = reader.readLine();
            while (aLine != null) {
                Text text = Text.valueOf(aLine);
                if (text.length() > 35 && text.trim().toUpperCase().startsWith(SHABPCODE))
                    break;
                aLine = reader.readLine();
            }

            if (aLine == null)
                return NO;

        } finally {
            input.reset();
        }

        return YES;
    }

    /**
     * Returns false. This class can write data to an S/HABP print file.
     */
    @Override
    public boolean canWriteData() {
        return false;
    }

    /**
     * Compares this object with the specified object for order based on the
     * <code>toString().compareTo(o.toString())</code> method. Returns a negative integer,
     * zero, or a positive integer as this object is less than, equal to, or greater than
     * the specified object.
     */
    @Override
    public int compareTo(FTableReader o) {
        requireNonNull(o, MessageFormat.format(RESOURCES.getString("paramNullErr"), "o"));
        return this.toString().compareTo(o.toString());
    }

    /**
     * Method that reads in S/HABP print formatted data from the specified input stream
     * and returns that data as an {@link FTableList} object.
     *
     * @param input An input stream containing the POST table formatted data.
     * @return An {@link FTableList} object that contains the data read in from the
     *         specified stream.
     * @throws IOException If there is a problem reading the specified stream.
     */
    @Override
    public FTableList read(InputStream input) throws IOException {
        requireNonNull(input, MessageFormat.format(RESOURCES.getString("paramNullErr"), "input"));

        //  S/HABP print files are required to be in ASCII with U.S. style number formatting.
        //  Get the default locale and save it.  Then change the default locale to US.
        Locale defLocale = Locale.getDefault();
        Locale.setDefault(Locale.US);

        try {
            // Create a reader to translate the input stream.
            LineNumberReader reader = new LineNumberReader(new InputStreamReader(input));

            FTableList db = readStream(reader);

            return db;

        } finally {
            //  Restore the default locale.
            Locale.setDefault(defLocale);
        }
    }

    /**
     * Always throws an exception since this class does not support writing.
     *
     * @param output The output stream to which the data is to be written.
     * @param tables The list of <code>Float64Table</code> objects to be written out.
     * @throws IOException Always thrown since this class does not support writing.
     */
    @Override
    public void write(OutputStream output, FTableList tables) throws IOException {
        throw new IOException(format(RESOURCES.getString("writeNotImplementedMsg"), "S/HABP"));
    }

    /**
     * Read in an S/HABP print formatted multi-table stream and create a new table
     * database with the data.
     *
     * @param in A file reader that reads in the characters from the file.
     * @return A table database containing the tables read in from the file.
     */
    private static FTableList readStream(LineNumberReader in) throws IOException {

        // Create an empty table database.
        FTableList db = FTableList.newInstance();

        try {
            //  A list of all the cases read from the file.
            FastTable<FastTable<FastTable<Float64>>> allCases = FastTable.newInstance();

            //  Read in the first case.
            FastTable<FastTable<Float64>> aCase = FastTable.newInstance();
            FastTable<Float64> machNumbers = FastTable.newInstance();
            Float64 caseMach = readCase(in, aCase);
            if (caseMach == null || aCase.isEmpty())
                throw new IOException(RESOURCES.getString("shabpNoDataErr"));
            allCases.add(aCase);
            machNumbers.add(caseMach);

            //  Read until we run out of cases.
            aCase = FastTable.newInstance();
            caseMach = readCase(in, aCase);
            while (caseMach != null && !aCase.isEmpty()) {
                allCases.add(aCase);
                machNumbers.add(caseMach);

                //  Look for the next case.
                aCase = FastTable.newInstance();
                caseMach = readCase(in, aCase);
            }

            if (allCases.size() == 1) {
                //  We have a set of 1-dimensional tables (in either alpha or beta).
                create1DTables(allCases.get(0), db);

            } else {
                //  We have either 2 or 3D tables in any combination of alpha, beta, and mach.
                create2Or3DTables(allCases, machNumbers, db);

            }

        } catch (NoSuchElementException | NumberFormatException e) {
            throw new IOException(format(RESOURCES.getString("parseErrMsg"), in.getLineNumber()), e);
        }

        return db;
    }

    /**
     * Method that reads in a single set of aero tables from the print file. This is the
     * data associated with a single analysis case.
     */
    private static Float64 readCase(LineNumberReader in, FastTable<FastTable<Float64>> caseData)
            throws IOException, NoSuchElementException, NumberFormatException {
        caseData.clear();

        //  Look for the "*** PRESSURE CALCULATION PROGRAM" sequence.
        String aLine = readToText(in, "*** PRESSURE CALCULATION PROGRAM");
        if (aLine == null)
            return null;        //  Couldn't find any more data.

        //  Read until we get to the next "Mach   Velocity" line.
        aLine = readToText(in, "Mach   Velocity");
        if (aLine == null)
            throw new IOException(RESOURCES.getString("shabpMissingMachLineErr"));
        in.readLine();
        aLine = in.readLine();
        if (aLine == null)
            throw new IOException(format(RESOURCES.getString("EOFErrMsg"), in.getLineNumber()));

        //  Parse out the Mach number.
        TextTokenizer tokenizer = TextTokenizer.newInstance();
        tokenizer.setDelimiters(DELIMITERS);
        tokenizer.setText(aLine);
        if (tokenizer.countTokens() < 1)
            throw new IOException(format(RESOURCES.getString("shabpMissingMachErr"), in.getLineNumber()));
        Float64 machNumber = Float64.valueOf(tokenizer.nextToken());

        //  Skip 7 lines and read in the 8th.
        for (int i = 0; i < 8; ++i)
            aLine = in.readLine();
        if (aLine == null)
            throw new IOException(format(RESOURCES.getString("EOFErrMsg"), in.getLineNumber()));

        //  Create some lists to hold the data.
        for (int i = 0; i < 8; ++i) {
            FastTable<Float64> tbl = FastTable.newInstance();
            caseData.add(tbl);
        }

        //  Begin reading in the table of data.
        boolean firstLine = true;
        while (aLine != null && !aLine.trim().equals("*** AERODYNAMIC ANALYSIS IS COMPLETE")) {
            //  Put the text in the tokenizer.
            tokenizer.setText(aLine);

            if (firstLine) {
                firstLine = false;

                //  Should be 8 tokens.
                if (tokenizer.countTokens() != 8)
                    throw new IOException(format(RESOURCES.getString("shabpWrongNuMColumnsErr"), in.getLineNumber()));

                //  Read in alpha, beta, CN, CA and CMb values.
                for (int i = 0; i < 5; ++i) {
                    Float64 value = Float64.valueOf(tokenizer.nextToken());
                    caseData.get(i).add(value);
                }

                //  Skip the remainder of this line (CL, CD and L/D).
            } else {
                firstLine = true;

                //  Should be 3 tokens.
                if (tokenizer.countTokens() != 3)
                    throw new IOException(format(RESOURCES.getString("shabpWrongNuMColumnsErr"), in.getLineNumber()));

                //  Read the CYb, CLNb, and CLLb values.
                for (int i = 5; i < 8; ++i) {
                    Float64 value = Float64.valueOf(tokenizer.nextToken());
                    caseData.get(i).add(value);
                }

            }

            //  Read in the next line.
            aLine = in.readLine();

        }   //  end while (!aLine.trim().equals("*** AERODYNAMIC ANALYSIS IS COMPLETE"))

        TextTokenizer.recycle(tokenizer);

        return machNumber;
    }

    /**
     * Reads lines of text from the specified reader until a line is found that starts
     * with the specified text (after white space is removed).
     */
    private static String readToText(LineNumberReader reader, String target) throws IOException {

        String aLine = reader.readLine();
        while (aLine != null && !aLine.trim().startsWith(target))
            aLine = reader.readLine();

        return aLine;
    }

    /**
     * Turns the data from a single case into a set of 1D tables.
     *
     * @param aCase A list of all the lists of values in a single case.
     * @param db    The table list we are adding the new tables to.
     */
    private static void create1DTables(FastTable<FastTable<Float64>> aCase, FTableList db) throws IOException {

        if (aCase.size() < 8)
            throw new IOException(format(RESOURCES.getString("shabpWrongNumTablesErr"), aCase.size()));

        //  Which variable is varying over this case, alpha or beta?
        boolean firstIndepAoA = true;
        FastTable<Float64> alphaList = aCase.get(0);
        FastTable<Float64> betaList = aCase.get(1);
        if (!betaList.get(0).equals(betaList.get(1))) {
            if (!alphaList.get(0).equals(alphaList.get(1)))
                throw new IOException(RESOURCES.getString("shabpCantDetAlphaBetaErr"));
            firstIndepAoA = false;      //  The 1st independent must be beta.
        }

        //  Create the array of independent values.
        String indepName = "ALPHA";
        Float64Vector indepValues;
        if (firstIndepAoA)
            indepValues = Float64Vector.valueOf(alphaList);
        else {
            indepName = "BETA";
            indepValues = Float64Vector.valueOf(betaList);
        }

        //  Create all the tables.
        for (int i = 2; i < 8; ++i) {
            String tblName = TBL_NAMES[i];
            Float64Vector depValues = Float64Vector.valueOf(aCase.get(i));
            Float64Table table = Float64Table.valueOf(tblName, indepName, indepValues, depValues);
            db.add(table);
        }
    }

    /**
     * Turns the data from multiple cases into a set of 2D or 3D tables.
     */
    private static void create2Or3DTables(FastTable<FastTable<FastTable<Float64>>> allCases, FastTable<Float64> allMachs, FTableList db) throws IOException {
        int numCases = allCases.size();

        //  Which variable is varying first, alpha or beta?
        boolean firstIndepAoA = true;
        FastTable<Float64> alphaList = allCases.get(0).get(0);
        //Float64 firstAlpha = alphaList.get(0);
        FastTable<Float64> betaList = allCases.get(0).get(1);
        Float64 firstBeta = betaList.get(0);
        if (!betaList.get(0).equals(betaList.get(1))) {
            if (!alphaList.get(0).equals(alphaList.get(1)))
                throw new IOException(RESOURCES.getString("shabpCantDetAlphaBetaErr"));
            firstIndepAoA = false;      //  The 1st independent must be beta.
        }

        //  Is Mach or the remaining aero angle varying 2nd?
        boolean secondIndepMach = false;
        Float64 firstMach = allMachs.get(0);
        if (!firstMach.equals(allMachs.get(1)))
            secondIndepMach = true;

        //  Are the tables 2D or 3D?
        boolean threeDTables = false;
        int num2ndBP = 0;
        if (secondIndepMach) {

            //  Check to see if the Mach number ever cycles.
            ++num2ndBP;
            for (int i = 1; i < numCases; ++i) {
                if (firstMach.equals(allMachs.get(i))) {
                    threeDTables = true;
                    break;
                }
                ++num2ndBP;
            }

        } else {
            //  Check to see if beta ever cycles.
            ++num2ndBP;
            for (int i = 1; i < numCases; ++i) {
                if (firstBeta.equals(allCases.get(i).get(1).get(0))) {
                    threeDTables = true;
                    break;
                }

                ++num2ndBP;
            }
        }

        if (DEBUG) {
            System.out.println("numCases = " + numCases + ", num2ndBP = " + num2ndBP);
            System.out.println("firstIndepAoA = " + firstIndepAoA + ", secondIndepMach = " + secondIndepMach);
            System.out.println("threeDTables = " + threeDTables);
            System.out.println("alphaList = " + alphaList);
            System.out.println("betaList = " + betaList);
            System.out.println("allMachs = " + allMachs);
        }

        int abIndex = 0;    //  2nd angle independent is alpha if 0 or beta if 1.
        if (firstIndepAoA)
            abIndex = 1;

        if (threeDTables) {
            //  Build up a set of 3D tables.
            FastTable<String> indepNames = FastTable.newInstance();
            Float64Vector indep1, indep2, indep3;

            FastTable<Float64> brkpoints = FastTable.newInstance();
            if (secondIndepMach) {  //  3rd independent is the remaining aero angle.
                //  Create Alpha/Beta breakpoints.
                indepNames.add(TBL_NAMES[abIndex]);
                for (int i = 0; i < numCases; i += num2ndBP)
                    brkpoints.add(allCases.get(i).get(abIndex).get(0));
                indep1 = Float64Vector.valueOf(brkpoints);

                //  Create Mach breakpoints.
                indepNames.add("MACH");
                brkpoints.clear();
                for (int i = 0; i < num2ndBP; ++i)
                    brkpoints.add(allMachs.get(i));
                indep2 = Float64Vector.valueOf(brkpoints);

            } else {    //  2nd independent is the remaining aero angle, 3rd is Mach.
                //  Create the Mach breakpoints.
                indepNames.add("MACH");
                for (int i = 0; i < numCases; i += num2ndBP)
                    brkpoints.add(allMachs.get(i));
                indep1 = Float64Vector.valueOf(brkpoints);

                //  Create Alpha/Beta breakpoints.
                indepNames.add(TBL_NAMES[abIndex]);
                brkpoints.clear();
                for (int i = 0; i < num2ndBP; ++i)
                    brkpoints.add(allCases.get(i).get(abIndex).get(0));
                indep2 = Float64Vector.valueOf(brkpoints);
            }

            if (firstIndepAoA) {
                indepNames.add(TBL_NAMES[0]);
                indep3 = Float64Vector.valueOf(alphaList);
            } else {
                indepNames.add(TBL_NAMES[1]);
                indep3 = Float64Vector.valueOf(betaList);
            }

            FastTable.recycle(brkpoints);

            //  Loop over all the tables to be created
            List<List<Float64Vector>> depData = FastTable.newInstance();
            for (int i = 2; i < 8; ++i) {
                String tblName = TBL_NAMES[i];

                //  Loop over all the cases 
                depData.clear();
                int num3rdBP = indep1.getDimension();
                int caseNum = 0;
                for (int j = 0; j < num3rdBP; ++j) {
                    List<Float64Vector> rows = FastTable.newInstance();
                    depData.add(rows);
                    for (int k = 0; k < num2ndBP; ++k, ++caseNum) {
                        List<Float64> column = allCases.get(caseNum).get(i);
                        rows.add(Float64Vector.valueOf(column));
                    }
                }

                //  Create the 3D table.
                Float64Table table = Float64Table.valueOf(tblName, indepNames, indep1, indep2, indep3, depData);
                db.add(table);

                //  Recycle the tables.
                for (List rows : depData)
                    FastTable.recycle((FastTable)rows);
            }   //  Next i

            //  Recycle the tables.
            FastTable.recycle((FastTable)depData);
            FastTable.recycle(indepNames);

        } else {
            //  Build up a set of 2D tables.
            FastTable<String> indepNames = FastTable.newInstance();
            Float64Vector indep1, indep2;

            FastTable<Float64> brkpoints = FastTable.newInstance();
            if (secondIndepMach) {
                //  Create Mach breakpoints.
                indepNames.add("MACH");
                for (int i = 0; i < num2ndBP; ++i)
                    brkpoints.add(allMachs.get(i));
                indep1 = Float64Vector.valueOf(brkpoints);

            } else {
                //  Create Alpha/Beta breakpoints.
                if (firstIndepAoA)
                    indepNames.add(TBL_NAMES[1]);
                else
                    indepNames.add(TBL_NAMES[0]);
                for (int i = 0; i < num2ndBP; ++i)
                    brkpoints.add(allCases.get(i).get(abIndex).get(0));
                indep1 = Float64Vector.valueOf(brkpoints);
            }

            if (firstIndepAoA) {
                indepNames.add(TBL_NAMES[0]);
                indep2 = Float64Vector.valueOf(alphaList);
            } else {
                indepNames.add(TBL_NAMES[1]);
                indep2 = Float64Vector.valueOf(betaList);
            }

            FastTable.recycle(brkpoints);

            //  Loop over all the tables to be created
            List<Float64Vector> depData = FastTable.newInstance();
            for (int i = 2; i < 8; ++i) {
                String tblName = TBL_NAMES[i];

                //  Loop over all the cases 
                depData.clear();
                int caseNum = 0;
                for (int k = 0; k < num2ndBP; ++k, ++caseNum) {
                    List<Float64> column = allCases.get(caseNum).get(i);
                    depData.add(Float64Vector.valueOf(column));
                }

                //  Create the 2D table.
                Float64Table table = Float64Table.valueOf(tblName, indepNames, indep1, indep2, depData);
                db.add(table);
            }   //  Next i

            //  Recycle the tables.
            FastTable.recycle((FastTable)depData);
            FastTable.recycle(indepNames);
        }

    }

}
